import asdf
import astropy
import pytest
from astropy import units as u
from astropy.cosmology import Planck15
from astropy.cosmology.units import with_H0
from astropy.units import equivalencies as eq
from packaging.version import Version


def create_equivalencies():
    result = [
        eq.plate_scale(0.3 * u.deg / u.mm),
        eq.pixel_scale(0.5 * u.deg / u.pix),
        eq.spectral_density(350 * u.nm, factor=2),
        eq.spectral_density(350 * u.nm),
        eq.spectral(),
        eq.brightness_temperature(500 * u.GHz),
        eq.brightness_temperature(500 * u.GHz, beam_area=23 * u.sr),
        with_H0(),
        eq.temperature_energy(),
        eq.temperature(),
        eq.thermodynamic_temperature(300 * u.Hz),
        eq.thermodynamic_temperature(140 * u.GHz, Planck15.Tcmb0),
        eq.beam_angular_area(3 * u.sr),
        eq.mass_energy(),
        eq.molar_mass_amu(),
        eq.doppler_relativistic(2 * u.m),
        eq.doppler_optical(2 * u.nm),
        eq.doppler_radio(2 * u.Hz),
        eq.parallax(),
        eq.logarithmic(),
        eq.dimensionless_angles(),
        eq.spectral() + eq.temperature(),
        (eq.spectral_density(35 * u.nm) + eq.brightness_temperature(5 * u.Hz, beam_area=2 * u.sr)),
        (eq.spectral() + eq.spectral_density(35 * u.nm) + eq.brightness_temperature(5 * u.Hz, beam_area=2 * u.sr)),
    ]

    if Version(astropy.__version__) >= Version("4.1"):
        result.append(eq.pixel_scale(100.0 * u.pix / u.cm))

    return result


@pytest.mark.parametrize("equivalency", create_equivalencies())
def test_serialization(equivalency, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["equivalency"] = equivalency
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert af["equivalency"] == equivalency



from packaging.version import parse as parse_version

from asdf_astropy.converters.helpers import parse_tag_version

from .core import TransformConverterBase, parameter_to_value


class ConstantConverter(TransformConverterBase):
    """
    ASDF support for serializing the Const1D and Const2D models.
    """

    # The 'dimensions' property was added in 1.4.0,
    # previously all values were 1D.
    _2D_MIN_VERSION = parse_version("1.4.0")

    tags = ("tag:stsci.edu:asdf/transform/constant-*",)
    types = (
        "astropy.modeling.functional_models.Const1D",
        "astropy.modeling.functional_models.Const2D",
    )

    def to_yaml_tree_transform(self, model, tag, ctx):
        from astropy.modeling.functional_models import Const1D, Const2D

        if parse_tag_version(tag) < self._2D_MIN_VERSION:
            if not isinstance(model, Const1D):
                msg = f"{tag} does not support models with > 1 dimension"
                raise TypeError(msg)

            return {"value": parameter_to_value(model.amplitude)}

        if isinstance(model, Const1D):
            dimension = 1
        elif isinstance(model, Const2D):
            dimension = 2

        return {"value": parameter_to_value(model.amplitude), "dimensions": dimension}

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling.functional_models import Const1D, Const2D

        if parse_tag_version(tag) < self._2D_MIN_VERSION:
            return Const1D(node["value"])

        if node["dimensions"] == 1:
            return Const1D(node["value"])

        if node["dimensions"] == 2:  # noqa: PLR2004
            return Const2D(node["value"])

        msg = f"Invalid dimensions: {node['dimensions']}"
        raise RuntimeError(msg)



import asdf
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import Angle, Latitude, Longitude


def create_angles():
    return [
        Angle(100, u.deg),
        Angle([100, 120, 150], u.deg),
        Angle([[90, 100, 110], [100, 120, 150]], u.deg),
        Angle(np.arange(100).reshape(5, 2, 10), u.deg),
        Latitude(10, u.deg),
        Longitude(-100, u.deg, wrap_angle=180 * u.deg),
    ]


@pytest.mark.parametrize("angle", create_angles())
def test_serialization(angle, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["angle"] = angle
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert (af["angle"] == angle).all()



import io
import sys
from pathlib import Path

import asdf
import pytest
import yaml


def test_resources():
    resources_root = Path(__file__).parent.parent / "resources"
    resource_manager = asdf.get_config().resource_manager

    for resource_path in resources_root.glob("**/*.yaml"):
        with resource_path.open("rb") as f:
            resource_content = f.read()
        resource = yaml.safe_load(resource_content)
        resource_uri = resource["id"]
        assert resource_manager[resource_uri] == resource_content


def test_manifests():
    manifests_root = Path(__file__).parent.parent / "resources" / "manifests"
    resource_manager = asdf.get_config().resource_manager

    for manifest_path in manifests_root.glob("*.yaml"):
        with manifest_path.open("rb") as f:
            manifest_content = f.read()
        manifest = yaml.safe_load(manifest_content)

        manifest_schema = asdf.schema.load_schema("asdf://asdf-format.org/core/schemas/extension_manifest-1.0.0")

        # The manifest must be valid against its own schema:
        asdf.schema.validate(manifest, schema=manifest_schema)

        for tag_definition in manifest["tags"]:
            # The tag's schema must be available:
            assert tag_definition["schema_uri"] in resource_manager


def test_extensions():
    package_and_uri_pairs = {(e.package_name, e.extension_uri) for e in asdf.get_config().extensions}

    assert ("asdf-astropy", "asdf://astropy.org/astropy/extensions/astropy-1.0.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.0.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.1.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.2.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.3.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.4.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.5.0") in package_and_uri_pairs


_ASTROPY_MODULES = [
    "astropy.coordinates",
    "astropy.io",
    "astropy.modeling",
    "astropy.table",
    "astropy.time",
    "astropy.units",
]


@pytest.fixture()
def _clean_astropy_imports():
    """Temporally unload all astropy modules used by asdf-astropy"""

    # If any astropy or asdf_astropy modules are already imported
    # remove them from sys.modules so we can later in this test check
    # if those modules were imported. We have to store these as
    # many submodules will reference or use types defined in other
    # submodules
    previous_modules = {}
    for name in sys.modules.copy():
        if name.startswith("asdf_astropy") or any(name.startswith(m) for m in _ASTROPY_MODULES):
            previous_modules[name] = sys.modules[name]
            del sys.modules[name]

    # Register a module finder that just raises an exception if
    # one of the tracked astropy modules is imported
    class _Finder:
        def find_spec(self, modulename, path=None, target=None):
            if any(modulename.startswith(m) for m in _ASTROPY_MODULES):
                msg = f"attempt to import astropy submodule({modulename}) during integration"
                raise Exception(msg)  # noqa: TRY002

    sys.meta_path.insert(0, _Finder())

    # Setup complete return the test
    yield

    # Restore the previously imported modules. This is necessary as other code
    # may have already used the modules we removed from the cache (astropy.constants
    # defines some quantities).
    for name in previous_modules:
        sys.modules[name] = previous_modules[name]
    if isinstance(sys.meta_path[0], _Finder):
        sys.meta_path.pop(0)


@pytest.mark.usefixtures("_clean_astropy_imports")
def test_no_astropy_import():
    """
    Confirm that none of the ASDF plugins import astropy modules
    at import time.
    """

    from asdf_astropy import integration

    integration.get_resource_mappings()
    integration.get_extensions()

    assert not any(k for k in sys.modules if any(k.startswith(m) for m in _ASTROPY_MODULES))


def test_no_core_extension_overwrite():
    """
    Check that this package (even though it implements converters for core types)
    does not overwrite the core extension provided by asdf
    """
    import astropy.units as u

    # define a tree with a unit that will be serialized by this package
    # the tree itself will be serialized by asdf (which should result in
    # including asdf as a used extension)
    tree = {"meter": u.m}

    bio = io.BytesIO()
    asdf.AsdfFile(tree).write_to(bio)
    bio.seek(0)

    with asdf.open(bio) as af:
        packages = [ext["software"]["name"] for ext in af.tree["history"]["extensions"]]
        # check that both asdf and asdf-astropy are recorded as being used to
        # generate the file
        assert "asdf" in packages
        assert "asdf-astropy" in packages



from .core import TransformConverterBase

__all__ = ["TabularConverter"]


class TabularConverter(TransformConverterBase):
    """
    ASDF support for serializing tabular models.
    """

    tags = ("tag:stsci.edu:asdf/transform/tabular-*",)
    types = (
        "astropy.modeling.tabular.Tabular1D",
        "astropy.modeling.tabular.Tabular2D",
    )

    def to_yaml_tree_transform(self, model, tag, ctx):
        node = {}
        if model.fill_value is not None:
            node["fill_value"] = model.fill_value
        node["lookup_table"] = model.lookup_table
        node["points"] = list(model.points)
        node["method"] = str(model.method)
        node["bounds_error"] = model.bounds_error

        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling import tabular

        lookup_table = node.pop("lookup_table")
        dim = lookup_table.ndim
        fill_value = node.pop("fill_value", None)
        if dim == 1:
            points = (node["points"][0],)
            model = tabular.Tabular1D(
                points=points,
                lookup_table=lookup_table,
                method=node["method"],
                bounds_error=node["bounds_error"],
                fill_value=fill_value,
            )
        elif dim == 2:  # noqa: PLR2004
            points = tuple(node["points"])
            model = tabular.Tabular2D(
                points=points,
                lookup_table=lookup_table,
                method=node["method"],
                bounds_error=node["bounds_error"],
                fill_value=fill_value,
            )
        else:
            msg = "tabular models with ndim > 2 are not supported "
            raise NotImplementedError(msg)

        return model



"""
Script that creates initial astropy manifest from the schemas
in the resources directory.  This file can be removed once
the format of the manifest files has been finalized.
"""
import argparse
from pathlib import Path

import yaml


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("schemas_path")
    parser.add_argument("output_path")
    return parser.parse_args()


class MultilineString(str):
    pass


def represent_multiline_string(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


if __name__ == "__main__":
    yaml.add_representer(MultilineString, represent_multiline_string)

    args = parse_args()

    manifest = {}

    manifest["id"] = "http://astropy.org/asdf/extensions/astropy/manifests/astropy-1.0"
    manifest["extension_uri"] = "http://astropy.org/asdf/extensions/astropy-1.0"
    manifest["title"] = "Astropy extension 1.0"
    manifest["description"] = MultilineString(
        "A set of tags for serializing astropy objects.  This does not include most\n"
        "model classes, which are handled by an implementation of the ASDF\n"
        "transform extension.",
    )
    manifest["asdf_standard_requirement"] = {
        "gte": "1.1.0",
    }
    manifest["tags"] = []

    for schema_path in Path(args.schemas_path).glob("**/*.yaml"):
        schema = yaml.safe_load(schema_path.read_bytes())
        tag_def = {
            "tag_uri": schema["id"].replace("http://astropy.org/schemas/astropy/", "tag:astropy.org:astropy/"),
            "schema_uri": schema["id"],
            "title": schema["title"].strip(),
        }

        if "tag" in schema:
            assert tag_def["tag_uri"] == schema["tag"]

        if "description" in schema:
            tag_def["description"] = MultilineString(schema["description"].strip())

        manifest["tags"].append(tag_def)

    with Path(args.output_path).open("w") as f:
        yaml.dump(manifest, f, sort_keys=False)



from asdf_astropy.converters.unit.quantity import QuantityConverter


class AngleConverter(QuantityConverter):
    tags = ("tag:astropy.org:astropy/coordinates/angle-*",)
    types = (
        "astropy.coordinates.angles.Angle",
        "astropy.coordinates.angles.core.Angle",
    )

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.angles import Angle

        return Angle(super().from_yaml_tree(node, tag, ctx))


class LatitudeConverter(QuantityConverter):
    tags = ("tag:astropy.org:astropy/coordinates/latitude-*",)
    types = (
        "astropy.coordinates.angles.Latitude",
        "astropy.coordinates.angles.core.Latitude",
    )

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.angles import Latitude

        return Latitude(super().from_yaml_tree(node, tag, ctx))


class LongitudeConverter(QuantityConverter):
    tags = ("tag:astropy.org:astropy/coordinates/longitude-*",)
    types = (
        "astropy.coordinates.angles.Longitude",
        "astropy.coordinates.angles.core.Longitude",
    )

    def to_yaml_tree(self, obj, tag, ctx):
        tree = super().to_yaml_tree(obj, tag, ctx)
        tree["wrap_angle"] = obj.wrap_angle
        return tree

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.angles import Longitude

        return Longitude(super().from_yaml_tree(node, tag, ctx), wrap_angle=node["wrap_angle"])



from asdf.extension import Converter
from astropy.utils import minversion


class ModelBoundingBoxConverter(Converter):
    tags = ("tag:stsci.edu:asdf/transform/property/bounding_box-1.0.0",)
    types = ("astropy.modeling.bounding_box.ModelBoundingBox",)

    def to_yaml_tree(self, bbox, tag, ctx):
        return {
            "intervals": {_input: list(interval) for _input, interval in bbox.named_intervals.items()},
            "ignore": list(bbox.ignored_inputs),
            "order": bbox.order,
        }

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.modeling.bounding_box import ModelBoundingBox, get_index, get_name

        intervals = {_input: tuple(interval) for _input, interval in node["intervals"].items()}

        ignored = node["ignore"] if "ignore" in node else []

        order = node["order"] if "order" in node else "C"

        def create_bounding_box(model, cbbox=None):
            if cbbox is None:
                ignore = ignored
            else:
                # Hack to pass compound_bounding_box selector_args ignore in 5.0.4+
                ignore = list(
                    set(ignored + [get_name(model, get_index(model, key)) for key in cbbox.selector_args.ignore]),
                )
                # Add in globally ignored inputs from the compound_bounding_box in 5.1+
                if minversion("astropy", "5.1"):
                    ignore = list(set(ignore + [get_name(model, get_index(model, key)) for key in cbbox.ignored]))

            return ModelBoundingBox(intervals, model, ignored=ignore, order=order)

        return create_bounding_box


class CompoundBoundingBoxConverter(Converter):
    tags = ("tag:stsci.edu:asdf/transform/property/compound_bounding_box-1.0.0",)
    types = ("astropy.modeling.bounding_box.CompoundBoundingBox",)

    def to_yaml_tree(self, cbbox, tag, ctx):
        node = {
            "selector_args": [{"argument": sa.name(cbbox._model), "ignore": sa.ignore} for sa in cbbox.selector_args],
            "cbbox": [{"key": list(key), "bbox": bbox} for key, bbox in cbbox.bounding_boxes.items()],
            "order": cbbox.order,
        }

        if minversion("astropy", "5.1"):
            node["ignore"] = cbbox.ignored_inputs

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.modeling.bounding_box import CompoundBoundingBox

        selector_args = tuple((selector["argument"], selector["ignore"]) for selector in node["selector_args"])
        bboxes = {tuple(bbox["key"]): bbox["bbox"] for bbox in node["cbbox"]}

        ignored = node["ignore"] if "ignore" in node else []

        order = node["order"] if "order" in node else "C"

        def create_bounding_box(model):
            if not minversion("astropy", "5.1"):
                if len(ignored) > 0:
                    msg = (
                        "Deserializing ignored elements of a compound bounding box is only supported for astropy 5.1+."
                    )
                    raise RuntimeError(msg)

                cbbox = CompoundBoundingBox({}, model, selector_args, order=order)
            else:
                cbbox = CompoundBoundingBox({}, model, selector_args, ignored=ignored, order=order)

            for key, bb in bboxes.items():
                cbbox[key] = bb(model, cbbox)

            return cbbox

        return create_bounding_box



__all__ = [
    "EquivalencyConverter",
    "QuantityConverter",
    "UnitConverter",
    "MagUnitConverter",
]

from .equivalency import EquivalencyConverter
from .magunit import MagUnitConverter
from .quantity import QuantityConverter
from .unit import UnitConverter



from asdf.extension import Converter


class EarthLocationConverter(Converter):
    tags = ("tag:astropy.org:astropy/coordinates/earthlocation-*",)
    types = ("astropy.coordinates.earth.EarthLocation",)

    def to_yaml_tree(self, obj, tag, ctx):
        return obj.info._represent_as_dict()

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.earth import EarthLocation

        return EarthLocation.info._construct_from_dict(node)



import unittest.mock as mk

import asdf
import pytest
from astropy.table import Table

from asdf_astropy.io.connect import read_table, write_table


def make_table():
    a = [1, 4, 5]
    b = [2.0, 5.0, 8.2]
    c = ["x", "y", "z"]
    return Table([a, b, c], names=("a", "b", "c"), meta={"name": "first table"})


def test_table_io(tmp_path):
    tmp_file = tmp_path / "table.asdf"

    table = make_table()

    table.write(tmp_file)

    # Simple sanity check using ASDF directly
    with asdf.open(tmp_file) as af:
        assert "data" in af
        assert isinstance(af["data"], Table)
        assert all(af["data"] == table)

    # Now test using the table reader
    new_t = Table.read(tmp_file)
    assert all(new_t == table)


def test_table_io_custom_key(tmp_path):
    tmp_file = tmp_path / "table.asdf"

    table = make_table()

    table.write(tmp_file, data_key="something")

    # Simple sanity check using ASDF directly
    with asdf.open(tmp_file) as af:
        assert "something" in af
        assert "data" not in af
        assert isinstance(af["something"], Table)
        assert all(af["something"] == table)

    # Now test using the table reader
    with pytest.raises(KeyError):
        new_t = Table.read(tmp_file)

    new_t = Table.read(tmp_file, data_key="something")
    assert all(new_t == table)


def test_table_io_custom_tree(tmp_path):
    tmp_file = tmp_path / "table.asdf"

    table = make_table()

    def make_custom_tree(tab):
        return {"foo": {"bar": tab}}

    table.write(tmp_file, make_tree=make_custom_tree)

    # Simple sanity check using ASDF directly
    with asdf.open(tmp_file) as af:
        assert "foo" in af
        assert "bar" in af["foo"]
        assert "data" not in af
        assert all(af["foo"]["bar"] == table)

    # Now test using table reader
    with pytest.raises(KeyError):
        new_t = Table.read(tmp_file)

    def find_table(asdffile):
        return asdffile["foo"]["bar"]

    new_t = Table.read(tmp_file, find_table=find_table)
    assert all(new_t == table)


def test_read_table_error(tmp_path):
    file_name = tmp_path / "table.asdf"

    with pytest.raises(ValueError, match="Options 'data_key' and 'find_table' are not compatible"):
        read_table(file_name, data_key=mk.MagicMock(), find_table=mk.MagicMock())


def test_write_table_error(tmp_path):
    file_name = tmp_path / "table.asdf"

    with pytest.raises(ValueError, match="Options 'data_key' and 'make_tree' are not compatible"):
        write_table(mk.MagicMock(), file_name, data_key=mk.MagicMock(), make_tree=mk.MagicMock())



import asdf
import numpy as np
import pytest
from asdf.testing.helpers import yaml_to_asdf
from astropy.io import fits
from numpy.testing import assert_array_equal

from asdf_astropy.testing.helpers import assert_hdu_list_equal


def create_hduls():
    hdul = fits.HDUList()
    header = fits.Header([("FOO", "BAR", "BAZ"), ("SOMENUM", "11.0"), ("EMPTY",)])
    hdul.append(fits.PrimaryHDU(header=header))
    hdul.append(fits.ImageHDU(data=np.arange(100)))

    hdul_with_table = fits.HDUList()
    hdul_with_table.append(fits.BinTableHDU.from_columns(np.array([(0, 1), (2, 3)], dtype=[("A", int), ("B", int)])))

    return [hdul, hdul_with_table]


@pytest.mark.parametrize("hdul", create_hduls())
def test_serialization(hdul, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["hdul"] = hdul
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_hdu_list_equal(af["hdul"], hdul)


def test_asdf_tag():
    yaml = """
hdul: !fits/fits-1.0.0
  - header:
      - [SIMPLE, true, conforms to FITS standard]
      - [BITPIX, 8, array data type]
      - [NAXIS, 0, number of array dimensions]
      - [EXTEND, true]
      - []
      - ['', Top Level MIRI Metadata]
      - []
      - [DATE, '2013-08-30T10:49:55.070373', The date this file was created (UTC)]
      - [FILENAME, MiriDarkReferenceModel_test.fits, The name of the file]
      - [TELESCOP, JWST, The telescope used to acquire the data]
      - []
      - ['', Information about the observation]
      - []
      - [DATE-OBS, '2013-08-30T10:49:55.000000', The date the observation was made (UTC)]
  - data: !core/ndarray-1.0.0
      data: [2, 3, 3, 4]
      datatype: float32
      shape: [4]
    header:
      - [XTENSION, IMAGE, Image extension]
      - [BITPIX, -32, array data type]
      - [NAXIS, 4, number of array dimensions]
      - [NAXIS1, 4]
      - [NAXIS2, 3]
      - [NAXIS3, 3]
      - [NAXIS4, 2]
      - [PCOUNT, 0, number of parameters]
      - [GCOUNT, 1, number of groups]
      - [EXTNAME, SCI, extension name]
      - [BUNIT, DN, Units of the data array]
  - data: !core/ndarray-1.0.0
      data: [5, 6, 7, 8]
      datatype: int64
      shape: [4]
    header:
      - [XTENSION, IMAGE, Image extension]
      - [BITPIX, -32, array data type]
      - [NAXIS, 4, number of array dimensions]
      - [NAXIS1, 4]
      - [NAXIS2, 3]
      - [NAXIS3, 3]
      - [NAXIS4, 2]
      - [PCOUNT, 0, number of parameters]
      - [GCOUNT, 1, number of groups]
      - [EXTNAME, ERR, extension name]
      - [BUNIT, DN, Units of the error array]
    """

    buff = yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        hdul = af["hdul"]
        assert len(hdul) == 3  # noqa: PLR2004
        assert hdul[0].header["FILENAME"] == "MiriDarkReferenceModel_test.fits"
        assert hdul[0].data is None
        assert hdul[1].header["EXTNAME"] == "SCI"
        assert_array_equal(hdul[1].data, np.array([2, 3, 3, 4], dtype=np.float32))
        assert hdul[2].header["EXTNAME"] == "ERR"
        assert_array_equal(hdul[2].data, np.array([5, 6, 7, 8], dtype=np.int64))



import unittest.mock as mk
from datetime import datetime

import asdf
import numpy as np
import pytest
from asdf.testing.helpers import yaml_to_asdf
from astropy import units as u
from astropy.coordinates import EarthLocation
from astropy.time import Time

from asdf_astropy.testing.helpers import assert_time_equal


def create_times():
    return [
        Time(1950.0, format="byear"),
        Time("B1950.0", format="byear_str"),
        Time(
            [1, 2],
            location=EarthLocation(x=[1, 2] * u.m, y=[3, 4] * u.m, z=[5, 6] * u.m),
            format="cxcsec",
        ),
        Time(datetime(2000, 1, 2, 12, 0, 0), format="datetime"),
        Time(2000.45, format="decimalyear"),
        Time("2000-01-01T00:00:00.000", format="fits"),
        Time(630720013.0, format="gps"),
        Time("2000-01-01 00:00:00.000", format="iso"),
        Time("2000-01-01T00:00:00.000", format="isot"),
        Time(2451544.5, format="jd"),
        Time(2000.0, format="jyear"),
        Time(
            "J2000.000",
            location=EarthLocation(x=6378100 * u.m, y=0 * u.m, z=0 * u.m),
            format="jyear_str",
        ),
        Time(51544.0, format="mjd"),
        Time(Time("2000-01-01T00:00:00.000").plot_date, format="plot_date"),  # astropy/astropy#14080
        Time(np.arange(100), format="unix"),
        Time(946684800.0, format="unix_tai"),
        Time("2000:001:00:00:00.000", format="yday"),
        Time("2000-01-01T00:00:00.000"),
        Time({"year": 2010, "month": 3, "day": 1}, format="ymdhms"),
        Time(np.datetime64("2000-01-01T01:01:01"), format="datetime64"),
        Time(["2001-01-02T12:34:56", "2001-02-03T00:01:02"]),
    ]


@pytest.mark.parametrize("time", create_times())
@pytest.mark.parametrize("version", asdf.versioning.supported_versions)
def test_serialization(time, version, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile(version=version) as af:
        af["time"] = time
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_time_equal(af["time"], time)


def create_examples():
    return [
        {"example": '!time/time-1.1.0 "2000-12-31T13:05:27.737"', "truth": Time("2000-12-31T13:05:27.737")},
        {"example": '!time/time-1.1.0 "2001:003:04:05:06.789"', "truth": Time("2001:003:04:05:06.789")},
        {"example": "!time/time-1.1.0 B2000.0", "truth": Time("B2000.0")},
        {
            "example": """!time/time-1.1.0
          value: 2000.0
          format: byear""",
            "truth": Time(2000.0, format="byear"),
        },
        {
            "example": """!time/time-1.1.0
          ["2000-12-31T13:05:27.737", "2000-12-31T13:06:38.444"]""",
            "truth": Time(["2000-12-31T13:05:27.737", "2000-12-31T13:06:38.444"]),
        },
        {
            "example": """!time/time-1.1.0
          value: !core/ndarray-1.0.0
            data: [2000, 2001]
            datatype: float64
          format: jyear""",
            "truth": Time([2000, 2001], format="jyear"),
        },
        {
            "example": """!time/time-1.1.0
          value: !core/ndarray-1.0.0
          value: 2000.0
          format: jyear
          scale: tdb
          location:
            x: !unit/quantity-1.1.0
              value: 6378100
              unit: !unit/unit-1.0.0 m
            y: !unit/quantity-1.1.0
              value: 0
              unit: !unit/unit-1.0.0 m
            z: !unit/quantity-1.1.0
              value: 0
              unit: !unit/unit-1.0.0 m
          format: jyear""",
            "truth": Time(
                2000.0,
                location=EarthLocation(x=6378100 * u.m, y=0 * u.m, z=0 * u.m),
                scale="tdb",
                format="jyear",
            ),
        },
    ]


@pytest.mark.parametrize("example", create_examples())
def test_read_examples(example):
    buff = yaml_to_asdf(f"example: {example['example'].strip()}")
    with asdf.AsdfFile() as af:
        af._open_impl(af, buff, mode="rw")
        assert np.all(af["example"] == example["truth"])


def test_error():
    from asdf_astropy.converters.time.time import TimeConverter

    with mk.patch("asdf_astropy.converters.time.time._ASTROPY_FORMAT_TO_ASDF_FORMAT") as mock_time:
        mock_time.get.return_value = "Bad time"

        example = "2000-01-01 00:00:00.000"
        with pytest.raises(ValueError, match=f"ASDF time '{example}' is not one of the recognized implicit formats"):
            TimeConverter().from_yaml_tree(example, mk.MagicMock(), mk.MagicMock())


def create_formats():
    from astropy.time.formats import TIME_FORMATS

    formats = []
    for format_ in TIME_FORMATS:
        new = Time("B2000.0")
        new.format = format_
        formats.append(new)

    return formats


@pytest.mark.parametrize("time", create_formats())
def test_formats(time, tmp_path):
    atol = 1 * u.day if "jyear" in time.format else 1e-08 * u.day

    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["time"] = time
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert af["time"].isclose(time, atol=atol)
        assert af["time"].format == time.format


def create_str_formats():
    return [
        Time("2000-01-01 00:00:00.000", format="iso"),
        Time("B2000.0", scale="utc", format="byear_str"),
        Time("J2000.000", scale="utc", format="jyear_str"),
        Time("2000:001:00:00:00.000", format="yday"),
    ]


@pytest.mark.parametrize("time", create_str_formats())
def test_str_format(time, tmp_path):
    """Test times that should serialize as strings"""

    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        af["time"] = time
        af.write_to(file_path)

    with asdf.open(file_path, _force_raw_types=True) as af:
        assert isinstance(af["time"], str)
        assert str(af["time"]) == time.value

    atol = 1 * u.day if "jyear" in time.format or "byear" in time.format else 1e-08 * u.day

    with asdf.open(file_path) as af:
        assert af["time"].isclose(time, atol=atol)
        assert af["time"].format == time.format



import abc

from asdf.extension import Converter

from asdf_astropy.converters.utils import import_type


def parameter_to_value(param):
    """
    Convert a model parameter to a Quantity or number,
    depending on the presence of a unit.

    Parameters
    ----------
    param : astropy.modeling.Parameter

    Returns
    -------
    astropy.units.Quantity or float
    """
    from astropy import units as u

    if param.unit is not None:
        return u.Quantity(param)

    return param.value


# One converter, UnitsMappingConverter, does not inherit
# this class.  When adding features here consider also
# updating UnitsMappingConverter.
# This class is used by other packages, e.g., gwcs, to implement
# converters for custom models.  Keep that in mind when modifying
# this code.
class TransformConverterBase(Converter):
    """
    ABC for transform/model converters.  Handles common
    properties after concrete converter sets model-specific
    properties.
    """

    @abc.abstractmethod
    def to_yaml_tree_transform(self, model, tag, ctx):
        """
        Convert a model's parameters into a dict suitable
        for ASDF serialization.  Common model properties
        such as name and inverse will be handled by this
        base class.

        Parameters
        ----------
        model : astropy.modeling.Model
            The model instance to convert.
        tag : str
            The tag identifying the YAML type that `astropy.modeling.Model` should be
            converted into.
        ctx : asdf.asdf.SerializationContext
            The context of the current serialization request.

        Returns
        -------
        dict
            ASDF node.
        """

    @abc.abstractmethod
    def from_yaml_tree_transform(self, node, tag, ctx):
        """
        Convert an ASDF node into an instance of the appropriate
        model class.  The implementing class need only instantiate
        the model and set parameter values; common model properties
        such as name and inverse will be handled by this base class.

        Parameters
        ----------
        node : dict
            The ASDF node to convert.
        tag : str
            The tag identifying the YAML type of the node.
        ctx : asdf.asdf.SerializationContext
            The context of the current serialization request.

        Returns
        -------
        astropy.modeling.Model
            The resulting model instance.
        """

    def to_yaml_tree(self, model, tag, ctx):
        from astropy.modeling.core import CompoundModel

        node = self.to_yaml_tree_transform(model, tag, ctx)

        if model.name is not None:
            node["name"] = model.name

        node["inputs"] = list(model.inputs)
        node["outputs"] = list(model.outputs)

        # Don't bother serializing analytic inverses provided
        # by the model:
        if getattr(model, "_user_inverse", None) is not None:
            node["inverse"] = model._user_inverse

        self._serialize_bounding_box(model, node)

        # model / parameter constraints
        if not isinstance(model, CompoundModel):
            fixed_nondefaults = {k: f for k, f in model.fixed.items() if f}
            if fixed_nondefaults:
                node["fixed"] = fixed_nondefaults
            bounds_nondefaults = {k: b for k, b in model.bounds.items() if any(b)}
            if bounds_nondefaults:
                node["bounds"] = bounds_nondefaults

        # model input_units_equivalencies
        if not isinstance(model, CompoundModel) and model.input_units_equivalencies:
            node["input_units_equivalencies"] = model.input_units_equivalencies

        return node

    def _serialize_bounding_box(self, model, node):
        from astropy.modeling.bounding_box import CompoundBoundingBox, ModelBoundingBox

        # ignore any default bounding_box
        if (bbox := model._user_bounding_box) is not None:
            if isinstance(bbox, ModelBoundingBox):
                self._serialize_bbox(model, node)
            elif isinstance(bbox, CompoundBoundingBox):
                self._serialize_cbbox(model, node)

    def _serialize_bbox(self, model, node):
        from astropy.modeling.bounding_box import ModelBoundingBox
        from astropy.utils import minversion

        bbox = model.bounding_box

        if minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
            if len(bbox.ignored) > 0:
                if minversion("astropy", "5.1"):
                    kwargs = {"_preserve_ignore": True}
                else:
                    msg = "Bounding box ignored arguments are only supported by astropy 5.1+"
                    raise RuntimeError(msg)
            else:
                kwargs = {}

            bbox = ModelBoundingBox.validate(model, bbox, **kwargs)
        else:
            if len(bbox.ignored) > 0:
                msg = "asdf-transform-schemas > 0.2.2 in order to serialize a bounding_box with ignored"
                raise RuntimeError(msg)

            bbox = bbox.bounding_box(order="C")
            bbox = list(bbox) if model.n_inputs == 1 else [list(item) for item in bbox]

        node["bounding_box"] = bbox

    def _serialize_cbbox(self, model, node):
        from astropy.utils import minversion

        bbox = model.bounding_box

        if minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
            node["bounding_box"] = bbox
        else:
            msg = "asdf-transform-schemas > 0.2.2 in order to serialize a compound bounding_box"
            raise RuntimeError(msg)

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.modeling.core import CompoundModel

        model = self.from_yaml_tree_transform(node, tag, ctx)

        if "name" in node:
            model.name = node["name"]

        if "inputs" in node:
            model.inputs = tuple(node["inputs"])

        if "outputs" in node:
            model.outputs = tuple(node["outputs"])

        self._deserialize_bounding_box(model, node)

        param_and_model_constraints = {}
        for constraint in ["fixed", "bounds"]:
            if constraint in node:
                param_and_model_constraints[constraint] = node[constraint]
        model._initialize_constraints(param_and_model_constraints)

        # this still writes eqs. for compound, but operates on each sub model
        if "input_units_equivalencies" in node and not isinstance(model, CompoundModel):
            model.input_units_equivalencies = node["input_units_equivalencies"]

        yield model

        if "inverse" in node:
            model.inverse = node["inverse"]

    def _deserialize_bounding_box(self, model, node):
        if "bounding_box" in node:
            bounding_box = node["bounding_box"]

            if isinstance(bounding_box, list):
                model.bounding_box = bounding_box
            elif callable(bounding_box):
                model.bounding_box = bounding_box(model)
            else:
                msg = f"Cannot form bounding_box from: {bounding_box}"
                raise TypeError(msg)


class SimpleTransformConverter(TransformConverterBase):
    """
    Class for converters that serialize all of a model's parameters
    and do not require special behavior based on tag version.

    Parameters
    ----------
    tags : list of str
        Tag patterns.

    model_type_name
        Fully-qualified model type name.
    """

    def __init__(self, tags, model_type_name):
        self._tags = tags
        self._model_type_name = model_type_name
        self._model_type = None

    @property
    def tags(self):
        return self._tags

    @property
    def types(self):
        return [self._model_type_name]

    @property
    def model_type(self):
        # Delay import until the model class is needed to improve speed
        # of loading the extension.
        if self._model_type is None:
            self._model_type = import_type(self._model_type_name)
        return self._model_type

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {p: parameter_to_value(getattr(model, p)) for p in model.param_names}

    def from_yaml_tree_transform(self, node, tag, ctx):
        model_type = self.model_type
        model_kwargs = {}
        for param in model_type.param_names:
            if param in node:
                model_kwargs[param] = node[param]
        return model_type(**model_kwargs)



from asdf.extension import Converter


class RepresentationConverter(Converter):
    tags = ("tag:astropy.org:astropy/coordinates/representation-*",)
    types = (
        "astropy.coordinates.representation.CartesianDifferential",
        "astropy.coordinates.representation.CartesianRepresentation",
        "astropy.coordinates.representation.CylindricalDifferential",
        "astropy.coordinates.representation.CylindricalRepresentation",
        "astropy.coordinates.representation.PhysicsSphericalDifferential",
        "astropy.coordinates.representation.PhysicsSphericalRepresentation",
        "astropy.coordinates.representation.RadialDifferential",
        "astropy.coordinates.representation.RadialRepresentation",
        "astropy.coordinates.representation.SphericalCosLatDifferential",
        "astropy.coordinates.representation.SphericalDifferential",
        "astropy.coordinates.representation.SphericalRepresentation",
        "astropy.coordinates.representation.UnitSphericalCosLatDifferential",
        "astropy.coordinates.representation.UnitSphericalDifferential",
        "astropy.coordinates.representation.UnitSphericalRepresentation",
        # classes were moved in https://github.com/astropy/astropy/pull/14792
        "astropy.coordinates.representation.cartesian.CartesianDifferential",
        "astropy.coordinates.representation.cartesian.CartesianRepresentation",
        "astropy.coordinates.representation.cylindrical.CylindricalDifferential",
        "astropy.coordinates.representation.cylindrical.CylindricalRepresentation",
        "astropy.coordinates.representation.spherical.PhysicsSphericalDifferential",
        "astropy.coordinates.representation.spherical.PhysicsSphericalRepresentation",
        "astropy.coordinates.representation.spherical.RadialDifferential",
        "astropy.coordinates.representation.spherical.RadialRepresentation",
        "astropy.coordinates.representation.spherical.SphericalCosLatDifferential",
        "astropy.coordinates.representation.spherical.SphericalDifferential",
        "astropy.coordinates.representation.spherical.SphericalRepresentation",
        "astropy.coordinates.representation.spherical.UnitSphericalRepresentation",
        "astropy.coordinates.representation.spherical.UnitSphericalDifferential",
        "astropy.coordinates.representation.spherical.UnitSphericalCosLatDifferential",
    )

    def to_yaml_tree(self, obj, tag, ctx):
        components = {}
        for c in obj.components:
            value = getattr(obj, "_" + c, None)
            if value is not None:
                components[c] = value

        return {
            "type": type(obj).__name__,
            "components": components,
        }

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates import representation

        return getattr(representation, node["type"])(**node["components"])



import itertools
import unittest.mock as mk
import warnings

import asdf
import astropy
import astropy.modeling
import numpy as np
import pytest
from asdf.testing.helpers import yaml_to_asdf
from astropy import units as u
from astropy.modeling import models as astropy_models
from astropy.modeling.bounding_box import CompoundBoundingBox, ModelBoundingBox
from astropy.utils import minversion

from asdf_astropy import integration
from asdf_astropy.testing import helpers


def assert_bounding_box_roundtrip(bounding_box, tmp_path, version=None):
    message = (
        "asdf_astropy.converters.transforms.tests.test_transform.assert_bounding_box_round_trip is deprecated."
        "Use asdf_astropy.testing.helpers.assert_bounding_box_roundtrip instead."
    )
    warnings.warn(message, DeprecationWarning, stacklevel=2)

    return helpers.assert_bounding_box_roundtrip(bounding_box, tmp_path, version=version)


def assert_model_roundtrip(model, tmp_path, version=None):
    message = (
        "asdf_astropy.converters.transforms.tests.test_transform.assert_model_round_trip is deprecated."
        "Use asdf_astropy.testing.helpers.assert_model_roundtrip instead."
    )
    warnings.warn(message, DeprecationWarning, stacklevel=2)

    return helpers.assert_model_roundtrip(model, tmp_path, version=version)


@pytest.mark.skipif(
    not minversion("asdf_transform_schemas", "0.2.2", inclusive=False),
    reason="Schema not present until versions after asdf-transform-schemas 0.2.2",
)
def test_deprecations(tmp_path):
    # Test assert_bounding_box_roundtrip deprecation
    bbox = ModelBoundingBox((0, 1), astropy_models.Polynomial1D(1))
    with pytest.warns(DeprecationWarning, match=".*assert_bounding_box_roundtrip.*"):
        assert_bounding_box_roundtrip(bbox, tmp_path)

    # Test assert_model_roundtrip deprecation
    model = astropy_models.Gaussian2D()
    with pytest.warns(DeprecationWarning, match=".*assert_model_roundtrip.*"):
        assert_model_roundtrip(model, tmp_path)


def create_bounding_boxes():
    model_bounding_box = [
        ModelBoundingBox((0, 1), astropy_models.Polynomial1D(1)),
        ModelBoundingBox(((2, 3), (3, 4)), astropy_models.Polynomial2D(1)),
        ModelBoundingBox(((5, 6), (7, 8)), astropy_models.Polynomial2D(1), order="F"),
    ]

    # ignore option is not properly supported until astropy versions with the bug fix from:
    # astropy/astropy#13032 (milestone 5.0.5) is included
    if minversion("astropy", "5.0.4", inclusive=False):
        model_bounding_box.extend(
            [
                ModelBoundingBox((9, 10), astropy_models.Polynomial2D(1), ignored=["x"]),
                ModelBoundingBox((11, 12), astropy_models.Polynomial2D(1), ignored=["y"]),
            ],
        )

    compound_bounding_box = [
        CompoundBoundingBox({(1,): (0, 1), (2,): (2, 3)}, astropy_models.Polynomial2D(1), [("x", True)]),
        CompoundBoundingBox(
            {(1,): ((0, 1), (-1, 0)), (2,): ((2, 3), (-3, -2))},
            astropy_models.Polynomial2D(1),
            [("x", False)],
        ),
    ]
    if minversion("astropy", "5.1"):
        compound_bounding_box.extend(
            [
                CompoundBoundingBox(
                    {(1,): (0, 1), (2,): (2, 3)},
                    astropy_models.Polynomial2D(1),
                    [("x", False)],
                    ignored=["x"],
                ),
                CompoundBoundingBox(
                    {(1,): (0, 1), (2,): (2, 3)},
                    astropy_models.Polynomial2D(1),
                    [("x", False)],
                    ignored=["y"],
                ),
            ],
        )

    return model_bounding_box + compound_bounding_box


@pytest.mark.parametrize("bbox", create_bounding_boxes())
@pytest.mark.skipif(
    not minversion("asdf_transform_schemas", "0.2.2", inclusive=False),
    reason="Schema not present until versions after asdf-transform-schemas 0.2.2",
)
def test_round_trip_bounding_box(bbox, tmp_path):
    helpers.assert_bounding_box_roundtrip(bbox, tmp_path)


def create_single_models():  # noqa: PLR0915
    model_with_bounding_box = astropy_models.Shift(10)
    model_with_bounding_box.bounding_box = ((1, 7),)

    model_with_user_inverse = astropy_models.Shift(10)
    model_with_user_inverse.inverse = astropy_models.Shift(-7)

    model_with_constraints = astropy_models.Legendre2D(
        x_degree=1,
        y_degree=1,
        c0_0=1,
        c0_1=2,
        c1_0=3,
        fixed={"c1_0": True, "c0_1": True},
        bounds={"c0_0": (-10, 10)},
    )

    model_with_custom_inputs_outputs = astropy_models.Gaussian2D()
    model_with_custom_inputs_outputs.inputs = ("a", "b")
    model_with_custom_inputs_outputs.outputs = ("c",)

    result = [
        # Generic model features
        astropy_models.Shift(10, name="some model name"),
        model_with_bounding_box,
        model_with_user_inverse,
        model_with_constraints,
        model_with_custom_inputs_outputs,
        # astropy.modeling.functional_models
        astropy_models.AiryDisk2D(amplitude=10.0, x_0=0.5, y_0=1.5),
        astropy_models.Box1D(amplitude=10.0, x_0=0.5, width=5.0),
        astropy_models.Box2D(amplitude=10.0, x_0=0.5, x_width=5.0, y_0=1.5, y_width=7.0),
        astropy_models.Const1D(amplitude=5.0),
        astropy_models.Const2D(amplitude=5.0),
        astropy_models.Disk2D(amplitude=10.0, x_0=0.5, y_0=1.5, R_0=5.0),
        astropy_models.Ellipse2D(amplitude=10.0, x_0=0.5, y_0=1.5, a=2.0, b=4.0, theta=0.1),
        astropy_models.Exponential1D(amplitude=10.0, tau=3.5),
        astropy_models.Gaussian1D(amplitude=10.0, mean=5.0, stddev=3.0),
        astropy_models.Gaussian2D(amplitude=10.0, x_mean=5.0, y_mean=5.0, x_stddev=3.0, y_stddev=3.0),
        astropy_models.KingProjectedAnalytic1D(amplitude=10.0, r_core=5.0, r_tide=2.0),
        astropy_models.Linear1D(slope=2.0, intercept=1.5),
        astropy_models.Logarithmic1D(amplitude=10.0, tau=3.5),
        astropy_models.Lorentz1D(amplitude=10.0, x_0=0.5, fwhm=2.5),
        astropy_models.Moffat1D(amplitude=10.0, x_0=0.5, gamma=1.2, alpha=2.5),
        astropy_models.Moffat2D(amplitude=10.0, x_0=0.5, y_0=1.5, gamma=1.2, alpha=2.5),
        astropy_models.Multiply(3),
        astropy_models.Multiply(10 * u.m),
        astropy_models.Planar2D(slope_x=0.5, slope_y=1.2, intercept=2.5),
        astropy_models.RedshiftScaleFactor(z=2.5),
        astropy_models.RickerWavelet1D(amplitude=10.0, x_0=0.5, sigma=1.2),
        astropy_models.RickerWavelet2D(amplitude=10.0, x_0=0.5, y_0=1.5, sigma=1.2),
        astropy_models.Ring2D(amplitude=10.0, x_0=0.5, y_0=1.5, r_in=5.0, width=10.0),
        astropy_models.Scale(3.4),
        astropy_models.Sersic1D(amplitude=10.0, r_eff=1.0, n=4.0),
        astropy_models.Sersic2D(amplitude=10.0, r_eff=1.0, n=4.0, x_0=0.5, y_0=1.5, ellip=0.0, theta=0.0),
        astropy_models.Shift(2.0),
        astropy_models.Shift(2.0 * u.deg),
        astropy_models.Scale(3.4 * u.deg),
        astropy_models.Sine1D(amplitude=10.0, frequency=0.5, phase=1.0),
        astropy_models.Cosine1D(amplitude=10.0, frequency=0.5, phase=1.0),
        astropy_models.Tangent1D(amplitude=10.0, frequency=0.5, phase=1.0),
        astropy_models.ArcSine1D(amplitude=10.0, frequency=0.5, phase=1.0),
        astropy_models.ArcCosine1D(amplitude=10.0, frequency=0.5, phase=1.0),
        astropy_models.ArcTangent1D(amplitude=10.0, frequency=0.5, phase=1.0),
        astropy_models.Trapezoid1D(amplitude=10.0, x_0=0.5, width=5.0, slope=1.0),
        astropy_models.TrapezoidDisk2D(amplitude=10.0, x_0=0.5, y_0=1.5, R_0=5.0, slope=1.0),
        astropy_models.Voigt1D(x_0=0.55, amplitude_L=10.0, fwhm_L=0.5, fwhm_G=0.9),
        # astropy.modeling.mappings
        astropy_models.Identity(2),
        astropy_models.Mapping((0, 1), n_inputs=3),
        # astropy.modeling.math_functions
        astropy.modeling.math_functions.AbsoluteUfunc(),
        astropy.modeling.math_functions.AddUfunc(),
        astropy.modeling.math_functions.ArccosUfunc(),
        astropy.modeling.math_functions.ArccoshUfunc(),
        astropy.modeling.math_functions.ArcsinUfunc(),
        astropy.modeling.math_functions.ArcsinhUfunc(),
        astropy.modeling.math_functions.Arctan2Ufunc(),
        astropy.modeling.math_functions.ArctanUfunc(),
        astropy.modeling.math_functions.ArctanhUfunc(),
        astropy.modeling.math_functions.CbrtUfunc(),
        astropy.modeling.math_functions.CosUfunc(),
        astropy.modeling.math_functions.CoshUfunc(),
        astropy.modeling.math_functions.Deg2radUfunc(),
        astropy.modeling.math_functions.DivideUfunc(),
        astropy.modeling.math_functions.DivmodUfunc(),
        astropy.modeling.math_functions.Exp2Ufunc(),
        astropy.modeling.math_functions.ExpUfunc(),
        astropy.modeling.math_functions.Expm1Ufunc(),
        astropy.modeling.math_functions.FabsUfunc(),
        astropy.modeling.math_functions.Floor_divideUfunc(),
        astropy.modeling.math_functions.FmodUfunc(),
        astropy.modeling.math_functions.HypotUfunc(),
        astropy.modeling.math_functions.Log10Ufunc(),
        astropy.modeling.math_functions.Log1pUfunc(),
        astropy.modeling.math_functions.Log2Ufunc(),
        astropy.modeling.math_functions.LogUfunc(),
        astropy.modeling.math_functions.Logaddexp2Ufunc(),
        astropy.modeling.math_functions.LogaddexpUfunc(),
        astropy.modeling.math_functions.ModUfunc(),
        astropy.modeling.math_functions.MultiplyUfunc(),
        astropy.modeling.math_functions.NegativeUfunc(),
        astropy.modeling.math_functions.PositiveUfunc(),
        astropy.modeling.math_functions.PowerUfunc(),
        astropy.modeling.math_functions.Rad2degUfunc(),
        astropy.modeling.math_functions.ReciprocalUfunc(),
        astropy.modeling.math_functions.RemainderUfunc(),
        astropy.modeling.math_functions.RintUfunc(),
        astropy.modeling.math_functions.SinUfunc(),
        astropy.modeling.math_functions.SinhUfunc(),
        astropy.modeling.math_functions.SqrtUfunc(),
        astropy.modeling.math_functions.SquareUfunc(),
        astropy.modeling.math_functions.SubtractUfunc(),
        astropy.modeling.math_functions.TanUfunc(),
        astropy.modeling.math_functions.TanhUfunc(),
        astropy.modeling.math_functions.True_divideUfunc(),
        # astropy.modeling.physical_models
        astropy_models.BlackBody(scale=10.0, temperature=6000.0 * u.K),
        astropy_models.Drude1D(amplitude=10.0, x_0=0.5, fwhm=2.5),
        # TODO: NFW
        # astropy.modeling.polynomial
        astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5),
        astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5, domain=(0.0, 1.0), window=(1.5, 2.5)),
        astropy_models.Chebyshev2D(1, 1, c0_0=1, c0_1=2, c1_0=3),
        astropy_models.Chebyshev2D(
            1,
            1,
            c0_0=1,
            c0_1=2,
            c1_0=3,
            x_domain=(1.0, 2.0),
            y_domain=(3.0, 4.0),
            x_window=(5.0, 6.0),
            y_window=(7.0, 8.0),
        ),
        astropy_models.Hermite1D(2, c0=2, c1=3, c2=0.5),
        astropy_models.Hermite2D(1, 1, c0_0=1, c0_1=2, c1_0=3),
        astropy_models.Legendre1D(2, c0=2, c1=3, c2=0.5),
        astropy_models.Legendre2D(1, 1, c0_0=1, c0_1=2, c1_0=3),
        astropy_models.Polynomial1D(2, c0=1, c1=2, c2=3),
        astropy_models.Polynomial2D(1, c0_0=1, c0_1=2, c1_0=3),
        # astropy.modeling.spline
        astropy_models.Spline1D(
            np.array([-3.0, -3.0, -3.0, -3.0, -1.0, 0.0, 1.0, 3.0, 3.0, 3.0, 3.0]),
            np.array(
                [
                    0.10412331,
                    0.07013616,
                    -0.18799552,
                    1.35953147,
                    -0.15282581,
                    0.03923,
                    -0.04297299,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
            ),
            3,
        ),
        # astropy.modeling.powerlaws
        astropy_models.BrokenPowerLaw1D(amplitude=10, x_break=0.5, alpha_1=2.0, alpha_2=3.5),
        astropy_models.ExponentialCutoffPowerLaw1D(10, 0.5, 2.0, 7.0),
        astropy_models.LogParabola1D(
            amplitude=10,
            x_0=0.5,
            alpha=2.0,
            beta=3.0,
        ),
        astropy_models.PowerLaw1D(amplitude=10.0, x_0=0.5, alpha=2.0),
        astropy_models.SmoothlyBrokenPowerLaw1D(amplitude=10.0, x_break=5.0, alpha_1=2.0, alpha_2=3.0, delta=0.5),
        # astropy.modeling.projections
        astropy_models.AffineTransformation2D(
            matrix=np.array([[1.0, 2.0], [3.0, 4.0]]),
            translation=np.array([5.0, 6.0]),
        ),
        astropy_models.Pix2Sky_Airy(theta_b=75.8),
        astropy_models.Sky2Pix_Airy(theta_b=75.8),
        astropy_models.Pix2Sky_BonneEqualArea(theta1=44.3),
        astropy_models.Sky2Pix_BonneEqualArea(theta1=44.3),
        astropy_models.Pix2Sky_COBEQuadSphericalCube(),
        astropy_models.Sky2Pix_COBEQuadSphericalCube(),
        astropy_models.Pix2Sky_ConicEqualArea(sigma=89.5, delta=0.5),
        astropy_models.Sky2Pix_ConicEqualArea(sigma=89.5, delta=0.5),
        astropy_models.Pix2Sky_ConicEquidistant(sigma=89.5, delta=0.5),
        astropy_models.Sky2Pix_ConicEquidistant(sigma=89.5, delta=0.5),
        astropy_models.Pix2Sky_ConicOrthomorphic(sigma=88.0, delta=1.0),
        astropy_models.Sky2Pix_ConicOrthomorphic(sigma=88.0, delta=1.0),
        astropy_models.Pix2Sky_ConicPerspective(sigma=89.5, delta=0.5),
        astropy_models.Sky2Pix_ConicPerspective(sigma=89.5, delta=0.5),
        astropy_models.Pix2Sky_CylindricalEqualArea(lam=0.5),
        astropy_models.Sky2Pix_CylindricalEqualArea(lam=0.5),
        astropy_models.Pix2Sky_CylindricalPerspective(mu=1.5, lam=2.4),
        astropy_models.Sky2Pix_CylindricalPerspective(mu=1.5, lam=2.4),
        astropy_models.Pix2Sky_Gnomonic(),
        astropy_models.Sky2Pix_Gnomonic(),
        astropy_models.Pix2Sky_HEALPixPolar(),
        astropy_models.Sky2Pix_HEALPixPolar(),
        astropy_models.Pix2Sky_HEALPix(H=12.0, X=17.0),
        astropy_models.Sky2Pix_HEALPix(H=12.0, X=17.0),
        astropy_models.Pix2Sky_HammerAitoff(),
        astropy_models.Sky2Pix_HammerAitoff(),
        astropy_models.Pix2Sky_Mercator(),
        astropy_models.Sky2Pix_Mercator(),
        astropy_models.Pix2Sky_Molleweide(),
        astropy_models.Sky2Pix_Molleweide(),
        astropy_models.Pix2Sky_Parabolic(),
        astropy_models.Sky2Pix_Parabolic(),
        astropy_models.Pix2Sky_PlateCarree(),
        astropy_models.Sky2Pix_PlateCarree(),
        astropy_models.Pix2Sky_Polyconic(),
        astropy_models.Sky2Pix_Polyconic(),
        astropy_models.Pix2Sky_QuadSphericalCube(),
        astropy_models.Sky2Pix_QuadSphericalCube(),
        astropy_models.Pix2Sky_SansonFlamsteed(),
        astropy_models.Sky2Pix_SansonFlamsteed(),
        astropy_models.Pix2Sky_SlantOrthographic(xi=0.1, eta=0.2),
        astropy_models.Sky2Pix_SlantOrthographic(xi=0.1, eta=0.2),
        astropy_models.Pix2Sky_SlantZenithalPerspective(mu=1.5, phi0=15.0, theta0=80.0),
        astropy_models.Sky2Pix_SlantZenithalPerspective(mu=1.5, phi0=15.0, theta0=80.0),
        astropy_models.Pix2Sky_Stereographic(),
        astropy_models.Sky2Pix_Stereographic(),
        astropy_models.Pix2Sky_TangentialSphericalCube(),
        astropy_models.Sky2Pix_TangentialSphericalCube(),
        astropy_models.Pix2Sky_ZenithalEqualArea(),
        astropy_models.Sky2Pix_ZenithalEqualArea(),
        astropy_models.Pix2Sky_ZenithalEquidistant(),
        astropy_models.Sky2Pix_ZenithalEquidistant(),
        astropy_models.Pix2Sky_ZenithalPerspective(mu=1.5, gamma=15.0),
        astropy_models.Sky2Pix_ZenithalPerspective(mu=1.5, gamma=15.0),
        # astropy.modeling.rotations
        astropy_models.EulerAngleRotation(23, 14, 2.3, axes_order="xzx"),
        astropy_models.RotateCelestial2Native(5.63, -72.5, 180),
        astropy_models.RotateCelestial2Native(5.63 * u.deg, -72.5 * u.deg, 180 * u.deg),
        astropy_models.RotateNative2Celestial(5.63, -72.5, 180),
        astropy_models.RotateNative2Celestial(5.63 * u.deg, -72.5 * u.deg, 180 * u.deg),
        astropy_models.Rotation2D(angle=1.51),
        astropy_models.RotationSequence3D([1.2, 2.3, 3.4, 0.3], "xyzx"),
        astropy_models.SphericalRotationSequence([1.2, 2.3, 3.4, 0.3], "xyzy"),
        # astropy.modeling.tabular
        astropy_models.Tabular1D(points=np.arange(0, 5), lookup_table=[1.0, 10, 2, 45, -3]),
        astropy_models.Tabular1D(points=np.arange(0, 5) * u.pix, lookup_table=[1.0, 10, 2, 45, -3] * u.nm),
        astropy_models.Tabular2D(
            points=([1, 2, 3], [1, 2, 3]),
            lookup_table=np.arange(0, 9).reshape(3, 3),
            bounds_error=False,
            fill_value=None,
            method="nearest",
        ),
        astropy_models.Tabular2D(
            points=([1, 2, 3], [1, 2, 3]) * u.pix,
            lookup_table=np.arange(0, 9).reshape(3, 3) * u.nm,
            bounds_error=False,
            fill_value=None,
            method="nearest",
        ),
    ]

    # Test case for model with a metaclass generated abstract bounding_box
    # where a custom bounding box is stored
    gaussian_1d = astropy_models.Gaussian1D(10, 1.5, 0.25)
    gaussian_1d.bounding_box = [7, 8]
    result.append(gaussian_1d)

    # compound model with bounding box
    model = astropy_models.Shift(1) & astropy_models.Shift(2)
    model.bounding_box = ((1, 2), (3, 4))
    result.append(model)

    # compound model with bounding box
    model = astropy_models.Shift(1) & astropy_models.Shift(2) & astropy_models.Shift(3)
    model.bounding_box = ((1, 2), (3, 4), (5, 6))
    result.append(model)

    if minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
        # model with compound bounding box
        model = astropy_models.Shift(1) & astropy_models.Scale(2) & astropy_models.Identity(1)
        model.inputs = ("x", "y", "slit_id")
        bounding_boxes = {
            (0,): ((-0.5, 1047.5), (-0.5, 2047.5)),
            (1,): ((-0.5, 3047.5), (-0.5, 4047.5)),
        }
        bounding_box = CompoundBoundingBox.validate(model, bounding_boxes, selector_args=[("slit_id", True)], order="F")
        model.bounding_box = bounding_box
        result.append(model)

        model = astropy_models.Shift(1) & astropy_models.Shift(2) & astropy_models.Shift(3)
        model.inputs = ("x", "y", "z")
        bounding_boxes = {
            (0,): (1.0, 2.0),
            (1,): (3.0, 4.0),
        }
        if minversion("astropy", "5.1"):
            bounding_box = CompoundBoundingBox.validate(
                model,
                bounding_boxes,
                selector_args=[("x", True)],
                ignored=["y"],
            )
            model.bounding_box = bounding_box
            result.append(model)

    result.append(astropy_models.Plummer1D(mass=10.0, r_plum=5.0))

    # models with input_units_equivalencies
    # 1D model
    m1 = astropy_models.Shift(1 * u.kg)
    m1.input_units_equivalencies = {"x": u.mass_energy()}

    # 2D model
    m2 = astropy_models.Const2D(10 * u.Hz)
    m2.input_units_equivalencies = {"x": u.dimensionless_angles(), "y": u.dimensionless_angles()}

    # 2D model with only one input equivalencies
    m3 = astropy_models.Const2D(10 * u.Hz)
    m3.input_units_equivalencies = {"x": u.dimensionless_angles()}

    # model using equivalency that has args using units
    m4 = astropy_models.PowerLaw1D(amplitude=1 * u.m, x_0=10 * u.pix, alpha=7)
    m4.input_units_equivalencies = {"x": u.equivalencies.pixel_scale(0.5 * u.arcsec / u.pix)}

    result.extend([m1, m2, m3, m4])

    # compound models with input_units_equivalencies
    m1 = astropy_models.Gaussian1D(10 * u.K, 11 * u.arcsec, 12 * u.arcsec)
    m1.input_units_equivalencies = {"x": u.parallax()}
    m2 = astropy_models.Gaussian1D(5 * u.s, 2 * u.K, 3 * u.K)
    m2.input_units_equivalencies = {"x": u.temperature()}

    result.extend([m1 | m2, m1 & m2, m1 + m2])

    # fix_inputs models with input_units_equivalencies
    m1 = astropy_models.Pix2Sky_TAN()
    m1.input_units_equivalencies = {"x": u.dimensionless_angles(), "y": u.dimensionless_angles()}
    m2 = astropy_models.Rotation2D()
    m = m1 | m2

    result.extend([astropy_models.fix_inputs(m, {"x": 45}), astropy_models.fix_inputs(m, {0: 45})])

    if minversion("astropy", "5.1") and minversion("asdf-transform-schemas", "0.2.3"):
        result.append(astropy_models.Schechter1D(phi_star=1.0, m_star=2.0, alpha=3.0))

    return result


UNSUPPORTED_MODELS = [
    # FITS-specific and deemed unworthy of ASDF serialization:
    astropy.modeling.polynomial.InverseSIP,
    astropy.modeling.polynomial.SIP,
    # Base classes which should not be directly supported:
    astropy.modeling.core.Model,
    astropy.modeling.math_functions._NPUfuncModel,
    astropy.modeling.polynomial.OrthoPolynomialBase,
    astropy.modeling.polynomial.PolynomialModel,
    astropy.modeling.projections.Conic,
    astropy.modeling.projections.Cylindrical,
    astropy.modeling.projections.HEALPix,
    astropy.modeling.projections.Pix2SkyProjection,
    astropy.modeling.projections.Projection,
    astropy.modeling.projections.PseudoConic,
    astropy.modeling.projections.PseudoCylindrical,
    astropy.modeling.projections.QuadCube,
    astropy.modeling.projections.Sky2PixProjection,
    astropy.modeling.projections.Zenithal,
    # https://github.com/astropy/asdf-astropy/issues/6
    astropy.modeling.physical_models.NFW,
]

if minversion("astropy", "5.1") and not minversion("asdf-transform-schemas", "0.2.3"):
    UNSUPPORTED_MODELS.append(astropy.modeling.powerlaws.Schechter1D)


# Model added in astropy 5.1 and schema added after asdf-transform 0.2.2
if minversion("astropy", "5.1") and not minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
    UNSUPPORTED_MODELS.append(astropy.modeling.powerlaws.Schechter1D)

if minversion("astropy", "6.0.dev"):
    UNSUPPORTED_MODELS.append(astropy.modeling.functional_models.GeneralSersic2D)


@pytest.mark.parametrize("model", create_single_models())
def test_single_model(tmp_path, model):
    helpers.assert_model_roundtrip(model, tmp_path)


def get_all_models():
    def _iterate_model_classes():
        for _key, value in itertools.chain(
            astropy_models.__dict__.items(),
            astropy.modeling.math_functions.__dict__.items(),
        ):
            if (
                isinstance(value, type)
                and issubclass(value, astropy.modeling.core.Model)
                and value not in UNSUPPORTED_MODELS
            ):
                yield value

    return list(_iterate_model_classes())


@pytest.mark.parametrize("model", get_all_models())
def test_all_models_supported(model):
    """
    Test that all model classes in astropy have serialization
    support implemented in this package.  If this test fails,
    file an issue on GitHub for each missing model and add
    the model to the UNSUPPORTED_MODELS list above with
    a link to the issue in a comment.
    """

    extensions = integration.get_extensions()
    extension_manager = asdf.extension.ExtensionManager(extensions)

    message = f"Missing support for model: {model.__module__}.{model.__qualname__}"
    assert extension_manager.handles_type(model), message


def test_legacy_const(tmp_path):
    with asdf.config_context() as config:
        config.remove_extension("asdf://asdf-format.org/transform/extensions/transform-1.5.0")

        model = astropy_models.Const1D(amplitude=5.0)
        helpers.assert_model_roundtrip(model, tmp_path, version="1.3.0")

        model = astropy_models.Const2D(amplitude=5.0)
        with pytest.raises(TypeError, match=r".* does not support models with > 1 dimension"):
            helpers.assert_model_roundtrip(model, tmp_path, version="1.3.0")


COMPOUND_OPERATORS = [
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__pow__",
    "__or__",
    "__and__",
]


@pytest.mark.parametrize("operator", COMPOUND_OPERATORS)
def test_compound_model(tmp_path, operator):
    left_model = astropy_models.Shift(5)
    right_model = astropy_models.Shift(-1)
    model = getattr(left_model, operator)(right_model)
    result = helpers.assert_model_roundtrip(model, tmp_path)
    helpers.assert_model_equal(result.left, left_model)
    helpers.assert_model_equal(result.right, right_model)
    assert result.op == model.op


def test_fix_inputs(tmp_path):
    model = astropy_models.Gaussian2D(1, 2, 3, 4, 5)
    fixed_model = astropy_models.fix_inputs(model, {"x": 2.5})
    result = helpers.assert_model_roundtrip(fixed_model, tmp_path)
    helpers.assert_model_equal(result.left, model)
    assert result.right == fixed_model.right
    assert result.op == fixed_model.op


def test_units_mapping(tmp_path):
    # Basic mapping between units:
    model = astropy_models.UnitsMapping(((u.m, u.dimensionless_unscaled),))
    model.name = "foo"
    result = helpers.assert_model_roundtrip(model, tmp_path)
    assert result.mapping == model.mapping

    # Remove units:
    model = astropy_models.UnitsMapping(((u.m, None),))
    result = helpers.assert_model_roundtrip(model, tmp_path)
    assert result.mapping == model.mapping

    # Change a model to accept any units:
    model = astropy_models.UnitsMapping(((None, u.m),))
    result = helpers.assert_model_roundtrip(model, tmp_path)
    assert result.mapping == model.mapping

    # With equivalencies:
    model = astropy_models.UnitsMapping(
        ((u.m, u.dimensionless_unscaled),),
        input_units_equivalencies={"x": u.equivalencies.spectral()},
    )
    result = helpers.assert_model_roundtrip(model, tmp_path)
    assert result.mapping == model.mapping

    # Allow dimensionless on all inputs:
    model = astropy_models.UnitsMapping(
        ((u.m, u.dimensionless_unscaled), (u.s, u.Hz)),
        input_units_allow_dimensionless=True,
    )
    result = helpers.assert_model_roundtrip(model, tmp_path)
    assert result.mapping == model.mapping

    # Allow dimensionless selectively:
    model = astropy_models.UnitsMapping(
        ((u.m, u.dimensionless_unscaled), (u.s, u.Hz)),
        input_units_allow_dimensionless={"x0": True, "x1": False},
    )
    result = helpers.assert_model_roundtrip(model, tmp_path)
    assert result.mapping == model.mapping


@pytest.mark.parametrize("standard_version", [v for v in asdf.versioning.supported_versions if v >= "1.4.0"])
@pytest.mark.parametrize(
    "model",
    [
        astropy_models.Polynomial1D(1, c0=5, c1=17),
        astropy_models.Polynomial1D(1, c0=5, c1=17, domain=[-5, 4], window=[-2, 3]),
        astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5, domain=[-2, 2]),
        astropy_models.Chebyshev1D(2, c0=2, c1=3, c2=0.5, domain=[-2, 2], window=[-0.5, 0.5]),
    ],
)
def test_1d_polynomial_with_asdf_standard_version(tmp_path, standard_version, model):
    result = helpers.assert_model_roundtrip(model, tmp_path, version=standard_version)
    assert result.domain == model.domain
    assert result.window == model.window


@pytest.mark.parametrize("standard_version", [v for v in asdf.versioning.supported_versions if v >= "1.4.0"])
@pytest.mark.parametrize(
    "model",
    [
        astropy_models.Polynomial2D(2, c0_0=3, c1_0=5, c0_1=7),
        astropy_models.Polynomial2D(
            2,
            c0_0=3,
            c1_0=5,
            c0_1=7,
            x_domain=[-2, 2],
            y_domain=[-4, 4],
            x_window=[-6, 6],
            y_window=[-8, 8],
        ),
        astropy_models.Chebyshev2D(1, 1, c0_0=1, c0_1=2, c1_0=3, x_domain=[-2, 2], y_domain=[-2, 2]),
        astropy_models.Chebyshev2D(
            1,
            1,
            c0_0=1,
            c0_1=2,
            c1_0=3,
            x_domain=[-2, 2],
            y_domain=[-2, 2],
            x_window=[-0.5, 0.5],
            y_window=[-0.1, 0.5],
        ),
    ],
)
def test_2d_polynomial_with_asdf_standard_version(tmp_path, standard_version, model):
    result = helpers.assert_model_roundtrip(model, tmp_path, version=standard_version)
    assert result.x_domain == model.x_domain
    assert result.y_domain == model.y_domain
    assert result.x_window == model.x_window
    assert result.y_window == model.y_window


def test_deserialize_compound_user_inverse(tmp_path):
    """
    Confirm that we are able to correctly reconstruct a
    compound model with a user inverse set on one of its
    component models.

    Due to code in TransformConverter that facilitates circular
    inverses, the user inverse of the component model is
    not available at the time that the CompoundModel is
    constructed.
    """

    yaml = """
model: !transform/concatenate-1.2.0
  forward:
  - !transform/shift-1.2.0
    inverse: !transform/shift-1.2.0 {offset: 5.0}
    offset: -10.0
  - !transform/shift-1.2.0 {offset: -20.0}
  """
    buff = yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        model = af["model"]
        assert model.has_inverse()
        assert model.inverse(-5, -20) == (0, 0)


def test_rotation_errors():
    from asdf_astropy.converters.transform.rotations import RotationSequenceConverter

    converter = RotationSequenceConverter()

    # to yaml error
    mdl = astropy_models.Const1D(5)
    mdl.angles = mk.MagicMock()
    mdl.axes_order = mk.MagicMock()
    with pytest.raises(TypeError, match=r"Cannot serialize model of type *"):
        converter.to_yaml_tree_transform(mdl, mk.MagicMock(), mk.MagicMock())

    # from yaml error
    node = {"angles": mk.MagicMock(), "axes_order": mk.MagicMock(), "rotation_type": mk.MagicMock()}
    with pytest.raises(ValueError, match=r"Unrecognized rotation_type: *"):
        converter.from_yaml_tree_transform(node, mk.MagicMock(), mk.MagicMock())


def test_projection_errors():
    from asdf_astropy.converters.transform.projections import ProjectionConverter

    converter = ProjectionConverter(mk.MagicMock(), mk.MagicMock(), mk.MagicMock())
    converter._sky2pix_type = astropy_models.Pix2Sky_Airy
    converter._pix2sky_type = astropy_models.Sky2Pix_Airy

    # to yaml error
    mdl = astropy_models.Const1D(5)
    with pytest.raises(TypeError, match=r"Unrecognized projection model type: *"):
        converter.to_yaml_tree_transform(mdl, mk.MagicMock(), mk.MagicMock())

    # from yaml error
    node = {"direction": mk.MagicMock()}
    with pytest.raises(ValueError, match=r"Unrecognized projection direction: *"):
        converter.from_yaml_tree_transform(node, mk.MagicMock(), mk.MagicMock())


def test_polynomial_errors():
    from asdf_astropy.converters.transform.polynomial import PolynomialConverter

    converter = PolynomialConverter()

    # from yaml error
    node = {"coefficients": np.zeros((2, 3))}
    with pytest.raises(TypeError, match=r"Coefficients must be an .* matrix"):
        converter.from_yaml_tree_transform(node, mk.MagicMock(), mk.MagicMock())


def test_compound_errors():
    from asdf_astropy.converters.transform.compound import CompoundConverter

    converter = CompoundConverter()

    # Left not a model
    tag = "!transform/add-1.2.0"
    node = {"forward": [mk.MagicMock(), mk.MagicMock()]}
    with pytest.raises(TypeError, match=r"Unknown left model type '.*'"):
        converter.from_yaml_tree_transform(node, tag, mk.MagicMock())

    # Right not a model (not fix_inputs)
    node = {"forward": [astropy_models.Const1D(17), mk.MagicMock()]}
    with pytest.raises(TypeError, match=r"Unknown right model type '.*'"):
        converter.from_yaml_tree_transform(node, tag, mk.MagicMock())

    # Right not a model (fix_inputs)
    tag = "!transform/fix_inputs-1.2.0"
    mdl = mk.MagicMock()
    mdl.__class__ = astropy_models.Const1D
    node = {"forward": [astropy_models.Const1D(17), mdl]}
    with pytest.raises(TypeError, match=r"Unknown right model type '.*'"):
        converter.from_yaml_tree_transform(node, tag, mk.MagicMock())


@pytest.mark.skipif(
    not minversion("asdf_transform_schemas", "0.2.2", inclusive=False),
    reason="Schema not present until versions after asdf-transform-schemas 0.2.2",
)
def test_bounding_box_missing_attributes():
    yaml = """
model: !transform/constant-1.4.0
    value: 1
    dimensions: 1
    bounding_box: !transform/property/bounding_box-1.0.0
        intervals:
            x: [1.0, 2.0]
    """
    buff = yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        model = af["model"]
        assert model.bounding_box.ignored == []
        assert model.bounding_box.order == "C"

    yaml = """
model: !transform/constant-1.4.0
    value: 1
    dimensions: 2
    bounding_box: !transform/property/compound_bounding_box-1.0.0
        selector_args:
            - argument: x
              ignore: true
        cbbox:
            - key: [0] # value of input x is 0 to select this box
              bbox: !transform/property/bounding_box-1.0.0
                intervals:
                    y: [1.0, 2.0]
            - key: [3] # value of input x is 3 to select this box
              bbox: !transform/property/bounding_box-1.0.0
                intervals:
                    y: [4.0, 5.0]
    """
    buff = yaml_to_asdf(yaml)
    with asdf.open(buff) as af:
        model = af["model"]
        assert model.bounding_box.order == "C"


@pytest.mark.filterwarnings("ignore:Unable to locate schema file for.*")
@pytest.mark.filterwarnings("ignore:.* not recognized, converting to raw Python data structure")
def test_compound_bbox_ignored_error():
    yaml = """
model: !transform/concatenate-1.2.0
  forward:
    - !transform/concatenate-1.2.0
      forward:
        - !transform/shift-1.2.0
          offset: 1.0
        - !transform/shift-1.2.0
          offset: 2.0
    - !transform/shift-1.2.0
      offset: 3.0
  bounding_box: !transform/property/compound_bounding_box-1.0.0
    selector_args:
      - argument: x
        ignore: true
    cbbox:
      - key: [0] # both value of input x is 0
        bbox: !transform/property/bounding_box-1.0.0
          intervals:
            x0: [2.0, 3.0]
      - key: [1] # both value of input x is 1
        bbox: !transform/property/bounding_box-1.0.0
          intervals:
            x0: [6.0, 7.0]
    ignore: [x1]
    """
    buff = yaml_to_asdf(yaml)

    if minversion("astropy", "5.1"):
        if not minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
            with pytest.raises(TypeError, match=r"Cannot form bounding_box from: *"):
                asdf.open(buff)
        else:
            asdf.open(buff)

    elif not minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
        with pytest.raises(TypeError, match=r"Cannot form bounding_box from: *"):
            asdf.open(buff)

    else:
        with pytest.raises(
            RuntimeError,
            match=r"Deserializing ignored elements of a compound bounding box is only supported for astropy 5.1+.",
        ):
            asdf.open(buff)


def test_serialize_bbox(tmp_path):
    mdl = astropy_models.Const2D(3)

    if minversion("astropy", "5.1"):
        from astropy.modeling import bind_bounding_box

        bind_bounding_box(mdl, (1, 2), ignored="y")
        if minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
            helpers.assert_model_roundtrip(mdl, tmp_path)
        else:
            with pytest.raises(
                RuntimeError,
                match=r"asdf-transform-schemas > 0.2.2 in order to serialize a bounding_box with ignored",
            ):
                helpers.assert_model_roundtrip(mdl, tmp_path)
    else:
        bbox = ModelBoundingBox({"x": (1, 2)}, mdl, ignored=["y"])
        mdl._user_bounding_box = bbox

        if minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
            with pytest.raises(
                RuntimeError,
                match=r"Bounding box ignored arguments are only supported by astropy 5.1+",
            ):
                helpers.assert_model_roundtrip(mdl, tmp_path)
        else:
            with pytest.raises(
                RuntimeError,
                match=r"asdf-transform-schemas > 0.2.2 in order to serialize a bounding_box with ignored",
            ):
                helpers.assert_model_roundtrip(mdl, tmp_path)


def test_serialize_cbbox(tmp_path):
    mdl = astropy_models.Shift(1) & astropy_models.Scale(2) & astropy_models.Identity(1)
    mdl.inputs = ("x", "y", "slit_id")
    bounding_boxes = {
        (0,): ((-0.5, 1047.5), (-0.5, 2047.5)),
        (1,): ((-0.5, 3047.5), (-0.5, 4047.5)),
    }
    bounding_box = CompoundBoundingBox.validate(mdl, bounding_boxes, selector_args=[("slit_id", True)], order="F")
    mdl.bounding_box = bounding_box

    if minversion("asdf_transform_schemas", "0.2.2", inclusive=False):
        helpers.assert_model_roundtrip(mdl, tmp_path)
    else:
        with pytest.raises(
            RuntimeError,
            match=r"asdf-transform-schemas > 0.2.2 in order to serialize a compound bounding_box",
        ):
            helpers.assert_model_roundtrip(mdl, tmp_path)



"""
This module builds all of the ASDF extensions which will be registered by `asdf_astropy.integration`,
via an ``entry-point`` in the ``pyproject.toml`` file.
"""

from asdf.extension import ManifestExtension
from astropy.utils import minversion

from ._manifest import CompoundManifestExtension
from .converters.coordinates.angle import AngleConverter, LatitudeConverter, LongitudeConverter
from .converters.coordinates.earth_location import EarthLocationConverter
from .converters.coordinates.frame import FrameConverter, LegacyICRSConverter
from .converters.coordinates.representation import RepresentationConverter
from .converters.coordinates.sky_coord import SkyCoordConverter
from .converters.coordinates.spectral_coord import SpectralCoordConverter
from .converters.fits.fits import AsdfFitsConverter, AstropyFitsConverter
from .converters.table.table import AsdfTableConverter, AstropyTableConverter, ColumnConverter, NdarrayMixinConverter
from .converters.time.time import TimeConverter
from .converters.time.time_delta import TimeDeltaConverter
from .converters.transform.compound import CompoundConverter
from .converters.transform.core import SimpleTransformConverter
from .converters.transform.functional_models import ConstantConverter
from .converters.transform.mappings import IdentityConverter, RemapAxesConverter, UnitsMappingConverter
from .converters.transform.math_functions import MathFunctionsConverter
from .converters.transform.polynomial import OrthoPolynomialConverter, PolynomialConverter
from .converters.transform.projections import ProjectionConverter
from .converters.transform.properties import CompoundBoundingBoxConverter, ModelBoundingBoxConverter
from .converters.transform.rotations import Rotate3DConverter, RotationSequenceConverter
from .converters.transform.spline import SplineConverter
from .converters.transform.tabular import TabularConverter
from .converters.unit.equivalency import EquivalencyConverter
from .converters.unit.magunit import MagUnitConverter
from .converters.unit.quantity import QuantityConverter
from .converters.unit.unit import UnitConverter

__all__ = [
    "TRANSFORM_CONVERTERS",
    "TRANSFORM_MANIFEST_URIS",
    "TRANSFORM_EXTENSIONS",
    "COORDINATES_CONVERTERS",
    "ASTROPY_CONVERTERS",
    "COORDINATES_EXTENSION",
    "ASTROPY_EXTENSIONS",
    "CORE_CONVERTERS",
    "CORE_MANIFEST_URIS",
    "CORE_EXTENSIONS",
]

TRANSFORM_CONVERTERS = [
    # astropy.modeling.core
    CompoundConverter(),
    # astropy.modeling.functional_models
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/airy_disk2d-*"],
        "astropy.modeling.functional_models.AiryDisk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/box1d-*"],
        "astropy.modeling.functional_models.Box1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/box2d-*"],
        "astropy.modeling.functional_models.Box2D",
    ),
    ConstantConverter(),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/disk2d-*"],
        "astropy.modeling.functional_models.Disk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ellipse2d-*"],
        "astropy.modeling.functional_models.Ellipse2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/exponential1d-*"],
        "astropy.modeling.functional_models.Exponential1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/gaussian1d-*"],
        "astropy.modeling.functional_models.Gaussian1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/gaussian2d-*"],
        "astropy.modeling.functional_models.Gaussian2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/king_projected_analytic1d-*"],
        "astropy.modeling.functional_models.KingProjectedAnalytic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/linear1d-*"],
        "astropy.modeling.functional_models.Linear1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/logarithmic1d-*"],
        "astropy.modeling.functional_models.Logarithmic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/lorentz1d-*"],
        "astropy.modeling.functional_models.Lorentz1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/moffat1d-*"],
        "astropy.modeling.functional_models.Moffat1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/moffat2d-*"],
        "astropy.modeling.functional_models.Moffat2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/multiplyscale-*"],
        "astropy.modeling.functional_models.Multiply",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/planar2d-*"],
        "astropy.modeling.functional_models.Planar2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/redshift_scale_factor-*"],
        "astropy.modeling.functional_models.RedshiftScaleFactor",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ricker_wavelet1d-*"],
        "astropy.modeling.functional_models.RickerWavelet1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ricker_wavelet2d-*"],
        "astropy.modeling.functional_models.RickerWavelet2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ring2d-*"],
        "astropy.modeling.functional_models.Ring2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/scale-*"],
        "astropy.modeling.functional_models.Scale",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sersic1d-*"],
        "astropy.modeling.functional_models.Sersic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sersic2d-*"],
        "astropy.modeling.functional_models.Sersic2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/shift-*"],
        "astropy.modeling.functional_models.Shift",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sine1d-*"],
        "astropy.modeling.functional_models.Sine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/cosine1d-*"],
        "astropy.modeling.functional_models.Cosine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/tangent1d-*"],
        "astropy.modeling.functional_models.Tangent1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/arcsine1d-*"],
        "astropy.modeling.functional_models.ArcSine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/arccosine1d-*"],
        "astropy.modeling.functional_models.ArcCosine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/arctangent1d-*"],
        "astropy.modeling.functional_models.ArcTangent1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/trapezoid1d-*"],
        "astropy.modeling.functional_models.Trapezoid1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/trapezoid_disk2d-*"],
        "astropy.modeling.functional_models.TrapezoidDisk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/voigt1d-*"],
        "astropy.modeling.functional_models.Voigt1D",
    ),
    # astropy.modeling.mappings
    IdentityConverter(),
    RemapAxesConverter(),
    # UnitsMapping is not represented here because
    # it is an astropy-specific transform and not
    # included in the ASDF transform extension.
    # astropy.modeling.math_functions
    MathFunctionsConverter(),
    # astropy.modeling.physical_models
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/blackbody-*"],
        "astropy.modeling.physical_models.BlackBody",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/drude1d-*"],
        "astropy.modeling.physical_models.Drude1D",
    ),
    # TODO: Implement NFW
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/plummer1d-*"],
        "astropy.modeling.physical_models.Plummer1D",
    ),
    # astropy.modeling.polynomial
    PolynomialConverter(),
    OrthoPolynomialConverter(),
    # SIP and InverseSIP are deliberately excluded because they
    # are FITS-specific and can be easily represented by a
    # simple combination of existing models.
    SplineConverter(),
    # astropy.modeling.powerlaws
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/broken_power_law1d-*"],
        "astropy.modeling.powerlaws.BrokenPowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/exponential_cutoff_power_law1d-*"],
        "astropy.modeling.powerlaws.ExponentialCutoffPowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/log_parabola1d-*"],
        "astropy.modeling.powerlaws.LogParabola1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/power_law1d-*"],
        "astropy.modeling.powerlaws.PowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/smoothly_broken_power_law1d-*"],
        "astropy.modeling.powerlaws.SmoothlyBrokenPowerLaw1D",
    ),
    # astropy.modeling.projections
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/affine-*"],
        "astropy.modeling.projections.AffineTransformation2D",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/airy-*"],
        "astropy.modeling.projections.Pix2Sky_Airy",
        "astropy.modeling.projections.Sky2Pix_Airy",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/bonne_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_BonneEqualArea",
        "astropy.modeling.projections.Sky2Pix_BonneEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cobe_quad_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_COBEQuadSphericalCube",
        "astropy.modeling.projections.Sky2Pix_COBEQuadSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_ConicEqualArea",
        "astropy.modeling.projections.Sky2Pix_ConicEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_equidistant-*"],
        "astropy.modeling.projections.Pix2Sky_ConicEquidistant",
        "astropy.modeling.projections.Sky2Pix_ConicEquidistant",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_orthomorphic-*"],
        "astropy.modeling.projections.Pix2Sky_ConicOrthomorphic",
        "astropy.modeling.projections.Sky2Pix_ConicOrthomorphic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_ConicPerspective",
        "astropy.modeling.projections.Sky2Pix_ConicPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cylindrical_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_CylindricalEqualArea",
        "astropy.modeling.projections.Sky2Pix_CylindricalEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cylindrical_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_CylindricalPerspective",
        "astropy.modeling.projections.Sky2Pix_CylindricalPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/gnomonic-*"],
        "astropy.modeling.projections.Pix2Sky_Gnomonic",
        "astropy.modeling.projections.Sky2Pix_Gnomonic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/hammer_aitoff-*"],
        "astropy.modeling.projections.Pix2Sky_HammerAitoff",
        "astropy.modeling.projections.Sky2Pix_HammerAitoff",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/healpix-*"],
        "astropy.modeling.projections.Pix2Sky_HEALPix",
        "astropy.modeling.projections.Sky2Pix_HEALPix",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/healpix_polar-*"],
        "astropy.modeling.projections.Pix2Sky_HEALPixPolar",
        "astropy.modeling.projections.Sky2Pix_HEALPixPolar",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/mercator-*"],
        "astropy.modeling.projections.Pix2Sky_Mercator",
        "astropy.modeling.projections.Sky2Pix_Mercator",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/molleweide-*"],
        "astropy.modeling.projections.Pix2Sky_Molleweide",
        "astropy.modeling.projections.Sky2Pix_Molleweide",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/parabolic-*"],
        "astropy.modeling.projections.Pix2Sky_Parabolic",
        "astropy.modeling.projections.Sky2Pix_Parabolic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/plate_carree-*"],
        "astropy.modeling.projections.Pix2Sky_PlateCarree",
        "astropy.modeling.projections.Sky2Pix_PlateCarree",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/polyconic-*"],
        "astropy.modeling.projections.Pix2Sky_Polyconic",
        "astropy.modeling.projections.Sky2Pix_Polyconic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/quad_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_QuadSphericalCube",
        "astropy.modeling.projections.Sky2Pix_QuadSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/sanson_flamsteed-*"],
        "astropy.modeling.projections.Pix2Sky_SansonFlamsteed",
        "astropy.modeling.projections.Sky2Pix_SansonFlamsteed",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/slant_orthographic-*"],
        "astropy.modeling.projections.Pix2Sky_SlantOrthographic",
        "astropy.modeling.projections.Sky2Pix_SlantOrthographic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/slant_zenithal_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_SlantZenithalPerspective",
        "astropy.modeling.projections.Sky2Pix_SlantZenithalPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/stereographic-*"],
        "astropy.modeling.projections.Pix2Sky_Stereographic",
        "astropy.modeling.projections.Sky2Pix_Stereographic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/tangential_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_TangentialSphericalCube",
        "astropy.modeling.projections.Sky2Pix_TangentialSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalEqualArea",
        "astropy.modeling.projections.Sky2Pix_ZenithalEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_equidistant-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalEquidistant",
        "astropy.modeling.projections.Sky2Pix_ZenithalEquidistant",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalPerspective",
        "astropy.modeling.projections.Sky2Pix_ZenithalPerspective",
    ),
    # astropy.modeling.rotations
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/rotate2d-*"],
        "astropy.modeling.rotations.Rotation2D",
    ),
    Rotate3DConverter(),
    RotationSequenceConverter(),
    # astropy.modeling.tabular
    TabularConverter(),
    # astropy.modeling.bounding_box
    ModelBoundingBoxConverter(),
    CompoundBoundingBoxConverter(),
]

if minversion("astropy", "5.1.0"):
    TRANSFORM_CONVERTERS.append(
        SimpleTransformConverter(
            ["tag:stsci.edu:asdf/transform/schechter1d-*"],
            "astropy.modeling.powerlaws.Schechter1D",
        ),
    )


# The order here is important; asdf will prefer to use extensions
# that occur earlier in the list.
TRANSFORM_MANIFEST_URIS = [
    "asdf://asdf-format.org/transform/manifests/transform-1.5.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.4.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.3.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.2.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.1.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.0.0",
]


TRANSFORM_EXTENSIONS = [
    ManifestExtension.from_uri(
        uri,
        # This prevents a warning about a missing extension when opening
        # files written by older versions of the asdf library:
        legacy_class_names=["astropy.io.misc.asdf.extension.AstropyAsdfExtension"],
        converters=TRANSFORM_CONVERTERS,
    )
    for uri in TRANSFORM_MANIFEST_URIS
]


COORDINATES_CONVERTERS = [
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/baseframe-*",
        "astropy.coordinates.baseframe.BaseCoordinateFrame",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/cirs-*",
        "astropy.coordinates.builtin_frames.cirs.CIRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk4-*",
        "astropy.coordinates.builtin_frames.fk4.FK4",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk4noeterms-*",
        "astropy.coordinates.builtin_frames.fk4.FK4NoETerms",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk5-*",
        "astropy.coordinates.builtin_frames.fk5.FK5",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/galactic-*",
        "astropy.coordinates.builtin_frames.galactic.Galactic",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/galactocentric-*",
        "astropy.coordinates.builtin_frames.galactocentric.Galactocentric",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/gcrs-*",
        "astropy.coordinates.builtin_frames.gcrs.GCRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/icrs-1.1.0",
        "astropy.coordinates.builtin_frames.icrs.ICRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/itrs-*",
        "astropy.coordinates.builtin_frames.itrs.ITRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/precessedgeocentric-*",
        "astropy.coordinates.builtin_frames.gcrs.PrecessedGeocentric",
    ),
    LegacyICRSConverter(),
    AngleConverter(),
    LatitudeConverter(),
    LongitudeConverter(),
    EarthLocationConverter(),
    RepresentationConverter(),
    SkyCoordConverter(),
    SpectralCoordConverter(),
]


ASTROPY_CONVERTERS = [
    UnitsMappingConverter(),
    TimeDeltaConverter(),
    AstropyTableConverter(),
    AstropyFitsConverter(),
    NdarrayMixinConverter(),
]


COORDINATES_EXTENSION = ManifestExtension.from_uri(
    "asdf://asdf-format.org/astronomy/coordinates/manifests/coordinates-1.0.0",
    converters=COORDINATES_CONVERTERS,
)


_ASTROPY_EXTENSION_MANIFEST_URIS = [
    "asdf://astropy.org/astropy/manifests/astropy-1.1.0",
    "asdf://astropy.org/astropy/manifests/astropy-1.0.0",
]


ASTROPY_EXTENSIONS = [
    ManifestExtension.from_uri(
        manifest_uri,
        # This prevents a warning about a missing extension when opening
        # files written by older versions of astropy:
        legacy_class_names=["astropy.io.misc.asdf.extension.AstropyExtension"],
        converters=ASTROPY_CONVERTERS,
    )
    for manifest_uri in _ASTROPY_EXTENSION_MANIFEST_URIS
]

# These tags are part of the ASDF Standard,
# but we want to override serialization here so that users can
# work with nice astropy objects for those entities.

CORE_CONVERTERS = [
    QuantityConverter(),
    TimeConverter(),
    ColumnConverter(),
    AsdfTableConverter(),
    AsdfFitsConverter(),
]

UNIT_CONVETERS = [
    UnitConverter(),
    EquivalencyConverter(),
    MagUnitConverter(),
]


CORE_MANIFEST_URIS = [
    "asdf://asdf-format.org/core/manifests/core-1.0.0",
    "asdf://asdf-format.org/core/manifests/core-1.1.0",
    "asdf://asdf-format.org/core/manifests/core-1.2.0",
    "asdf://asdf-format.org/core/manifests/core-1.3.0",
    "asdf://asdf-format.org/core/manifests/core-1.4.0",
    "asdf://asdf-format.org/core/manifests/core-1.5.0",
    "asdf://asdf-format.org/core/manifests/core-1.6.0",
]


CORE_EXTENSIONS = [
    CompoundManifestExtension(
        [
            ManifestExtension.from_uri(u, converters=CORE_CONVERTERS),
            ManifestExtension.from_uri("asdf://astropy.org/astropy/manifests/units-1.0.0", converters=UNIT_CONVETERS),
        ],
    )
    for u in CORE_MANIFEST_URIS
]



import asdf
import pytest
from astropy import units as u
from astropy.coordinates import EarthLocation, Latitude, Longitude
from astropy.coordinates.earth import ELLIPSOIDS
from astropy.units import Quantity

from asdf_astropy.testing.helpers import assert_earth_location_equal


def create_earth_locations():
    longitude = Longitude([0.0, 45.0, 90.0, 135.0, 180.0, -180, -90, -45], u.deg, wrap_angle=180 * u.deg)
    latitude = Latitude([+0.0, 30.0, 60.0, +90.0, -90.0, -60.0, -30.0, 0.0], u.deg)
    height = Quantity([0.1, 0.5, 1.0, -0.5, -1.0, +4.2, -11.0, -0.1], u.m)
    position = (longitude, latitude, height)

    result = [
        EarthLocation(lat=34.4900 * u.deg, lon=-104.221800 * u.deg, height=40 * u.km),
        EarthLocation(*EarthLocation.from_geodetic(*position).to_geocentric()),
    ]

    result.extend([EarthLocation.from_geodetic(*position, ellipsoid=e) for e in ELLIPSOIDS])

    return result


@pytest.mark.parametrize("earth_location", create_earth_locations())
def test_serialization(earth_location, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["earth_location"] = earth_location
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert (af["earth_location"] == earth_location).all()


@pytest.fixture()
def _builtin_site_registry():
    orig_sites = getattr(EarthLocation, "_site_registry", None)
    EarthLocation._get_site_registry(force_builtin=True)
    yield
    EarthLocation._site_registry = orig_sites


@pytest.mark.usefixtures("_builtin_site_registry")
def test_earthlocation_site(tmp_path):
    earth_location = EarthLocation.of_site("greenwich")

    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["earth_location"] = earth_location
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_earth_location_equal(af["earth_location"], earth_location)



import numpy as np
from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType

_GUESSABLE_FORMATS = {"iso", "byear", "jyear", "yday"}

_ASTROPY_FORMAT_TO_ASDF_FORMAT = {
    "isot": "iso",
    "byear_str": "byear",
    "jyear_str": "jyear",
}


class TimeConverter(Converter):
    tags = ("tag:stsci.edu:asdf/time/time-*",)
    types = ("astropy.time.core.Time",)

    def to_yaml_tree(self, obj, tag, ctx):
        from astropy.time import Time

        base_format = obj.format
        if base_format == "byear":
            obj = Time(obj, format="byear_str")
        elif base_format == "jyear":
            obj = Time(obj, format="jyear_str")
        elif base_format in ("fits", "datetime", "plot_date", "ymdhms", "datetime64"):
            obj = Time(obj, format="isot")

        asdf_format = _ASTROPY_FORMAT_TO_ASDF_FORMAT.get(obj.format, obj.format)
        guessable_format = asdf_format in _GUESSABLE_FORMATS

        if obj.scale == "utc" and guessable_format and obj.isscalar and base_format == obj.format:
            return obj.value

        node = {
            "value": obj.value,
        }

        if not guessable_format:
            node["format"] = asdf_format

        if base_format != obj.format:
            node["base_format"] = base_format

        if obj.scale != "utc":
            node["scale"] = obj.scale

        if obj.location is not None:
            # The 1.0.0 and 1.1.0 tags differ in how location is represented.
            # In 1.0.0, there is a single "unit" property that is shared among
            # x, y, and z, and in 1.1.0 each is a quantity with its own unit.
            location = (
                {
                    "x": obj.location.x.value,
                    "y": obj.location.y.value,
                    "z": obj.location.z.value,
                    "unit": obj.location.unit,
                }
                if tag.endswith("1.0.0")
                else {
                    "x": obj.location.x,
                    "y": obj.location.y,
                    "z": obj.location.z,
                }
            )

            node["location"] = location

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy import units
        from astropy.coordinates import EarthLocation
        from astropy.time import Time

        if isinstance(node, (str, list, np.ndarray, NDArrayType)):
            time = Time(node)
            asdf_format = _ASTROPY_FORMAT_TO_ASDF_FORMAT.get(time.format, time.format)
            if asdf_format not in _GUESSABLE_FORMATS:
                msg = f"ASDF time '{node}' is not one of the recognized implicit formats"
                raise ValueError(msg)

            return time

        location = node.get("location")
        if location is not None:
            # The 1.0.0 and 1.1.0 tags differ in how location is represented.
            # In 1.0.0, there is a single "unit" property that is shared among
            # x, y, and z, and in 1.1.0 each is a quantity with its own unit.
            if tag.endswith("1.0.0"):
                unit = location.get("unit", units.m)
                location = EarthLocation.from_geocentric(
                    units.Quantity(location["x"], unit=unit),
                    units.Quantity(location["y"], unit=unit),
                    units.Quantity(location["z"], unit=unit),
                )
            else:
                location = EarthLocation.from_geocentric(
                    location["x"],
                    location["y"],
                    location["z"],
                )

        time = Time(
            node["value"],
            format=node.get("format"),
            scale=node.get("scale"),
            location=location,
        )

        base_format = node.get("base_format")
        if base_format is not None and base_format != time.format:
            time.format = base_format

        return time



from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType


class ColumnConverter(Converter):
    tags = ("tag:stsci.edu:asdf/core/column-*",)
    types = (
        "astropy.table.column.Column",
        "astropy.table.column.MaskedColumn",
    )

    def to_yaml_tree(self, obj, tag, ctx):
        node = {"data": obj.data, "name": obj.name}

        if obj.description:
            node["description"] = obj.description

        if obj.unit:
            node["unit"] = obj.unit

        if obj.meta:
            node["meta"] = obj.meta

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.table import Column, MaskedColumn
        from numpy.ma.core import MaskedArray

        data = node["data"]
        if isinstance(data, NDArrayType):
            # TODO: Why doesn't NDArrayType work?  This needs some research
            # and documentation.
            data = data._make_array()

        column_class = MaskedColumn if isinstance(data, MaskedArray) else Column

        return column_class(
            data=data,
            name=node["name"],
            description=node.get("description"),
            unit=node.get("unit"),
            meta=node.get("meta"),
        )


class AsdfTableConverter(Converter):
    tags = ("tag:stsci.edu:asdf/core/table-*",)
    types = ()

    def to_yaml_tree(self, obj, tag, ctx):
        msg = "astropy does not support writing astropy.table.Table with the ASDF table-1.0.0 tag"
        raise NotImplementedError(msg)

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.table import Table

        return Table(node["columns"], meta=node.get("meta"))


class AstropyTableConverter(Converter):
    tags = ("tag:astropy.org:astropy/table/table-*",)
    types = (
        "astropy.table.table.Table",
        "astropy.table.table.QTable",
    )

    def to_yaml_tree(self, obj, tag, ctx):
        from astropy.table import QTable

        node = {
            "columns": [obj[name] for name in obj.colnames],
            "colnames": obj.colnames,
            "qtable": isinstance(obj, QTable),
        }

        if obj.meta:
            node["meta"] = obj.meta

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.table import QTable, Table

        table = QTable(meta=node.get("meta")) if node.get("qtable", False) else Table(meta=node.get("meta"))

        for name, column in zip(node["colnames"], node["columns"]):
            table[name] = column

        return table


class NdarrayMixinConverter(Converter):
    tags = ("tag:astropy.org:astropy/table/ndarraymixin-*",)
    types = ("astropy.table.ndarray_mixin.NdarrayMixin",)

    def to_yaml_tree(self, obj, tag, ctx):
        import numpy as np

        return {"array": np.asarray(obj)}

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.table import NdarrayMixin

        arr = node["array"]

        # this will trigger reading the ASDF block that contains the array data
        return arr.view(NdarrayMixin)



"""
Helpers for testing astropy objects in ASDF files.
"""
import numpy as np


def assert_earth_location_equal(a, b):
    """
    Assert earth location objects are equal.
    """
    __tracebackhide__ = True

    assert (a == b).all()


def assert_representation_equal(a, b):
    """
    Assert representation objects are equal.
    """
    from astropy import units

    __tracebackhide__ = True

    assert type(a) is type(b)
    assert a.components == b.components
    for component in a.components:
        assert units.allclose(getattr(a, component), getattr(b, component))


def assert_sky_coord_equal(a, b):
    """
    Assert sky coordinate objects are equal.
    """
    __tracebackhide__ = True

    assert a.is_equivalent_frame(b)
    assert a.representation_type is b.representation_type
    assert a.shape == b.shape

    assert_representation_equal(a.data, b.data)


def assert_frame_equal(a, b):
    """
    Assert frame objects are equal.
    """
    __tracebackhide__ = True

    assert type(a) is type(b)

    if a is None:
        return

    if a.has_data:
        assert b.has_data
        assert_representation_equal(a.data, b.data)
    else:
        return


def assert_spectral_coord_equal(a, b):
    """
    Assert spectral coordinate objects are equal.
    """
    from astropy.tests.helper import assert_quantity_allclose

    __tracebackhide__ = True

    assert type(a) is type(b)
    assert_quantity_allclose(a.quantity, b.quantity)
    assert_frame_equal(a.observer, b.observer)
    assert_frame_equal(a.target, b.target)


def assert_time_equal(a, b):
    """
    Assert time objects are equal
    """
    from astropy.coordinates import EarthLocation

    assert a.format == b.format
    assert a.scale == b.scale

    assert isinstance(a.location, type(b.location))
    if isinstance(a.location, EarthLocation):
        assert_earth_location_equal(a.location, b.location)
    else:
        assert a.location == b.location

    if a.format == "plot_date":
        np.testing.assert_array_almost_equal(a.value, b.value)
    else:
        np.testing.assert_array_equal(a, b)


def assert_time_delta_equal(a, b):
    """
    Assert time delta objects are equal
    """
    np.testing.assert_array_equal(a.jd, b.jd)
    np.testing.assert_array_equal(a.jd2, b.jd2)
    np.testing.assert_array_equal(a.sec, b.sec)


def assert_hdu_list_equal(a, b):
    """
    Assert fits hdulists are equal.
    """
    assert len(a) == len(b)
    for hdu_a, hdu_b in zip(a, b):
        np.testing.assert_array_equal(hdu_a.data, hdu_b.data)
        assert len(hdu_a.header.cards) == len(hdu_b.header.cards)
        for card_a, card_b in zip(hdu_a.header.cards, hdu_b.header.cards):
            assert tuple(card_a) == tuple(card_b)


def assert_description_equal(a, b):
    """
    Assert table descriptions are equal.
    """
    if a in ("", None) and b in ("", None):
        return

    assert a == b


def assert_table_equal(a, b):
    """
    Assert astropy tables are equal.
    """
    from astropy.table import Column, MaskedColumn

    assert type(a) == type(b)
    assert a.meta == b.meta

    assert len(a) == len(b)
    for row_a, row_b in zip(a, b):
        np.testing.assert_array_equal(row_a, row_b)

    assert a.colnames == b.colnames
    for column_name in a.colnames:
        col_a = a[column_name]
        col_b = b[column_name]
        if isinstance(col_a, (Column, MaskedColumn)) and isinstance(col_b, (Column, MaskedColumn)):
            assert_description_equal(col_a.description, col_b.description)
            assert col_a.unit == col_b.unit
            assert col_a.meta == col_b.meta
            np.testing.assert_array_equal(col_a.data, col_b.data)
            np.testing.assert_array_equal(
                getattr(col_a, "mask", [False] * len(col_a)),
                getattr(col_b, "mask", [False] * len(col_b)),
            )


def assert_table_roundtrip(table, tmp_path):
    """
    Assert that a table can be written to an ASDF file and read back
    in without losing any of its essential properties.
    """
    import asdf

    file_path = tmp_path / "testable.asdf"

    with asdf.AsdfFile({"table": table}) as af:
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_table_equal(table, af["table"])
        return af["table"]


def assert_model_equal(a, b):
    """
    Assert that two model instances are equivalent.
    """
    if a is None and b is None:
        return

    assert a.__class__ == b.__class__

    assert a.name == b.name
    assert a.inputs == b.inputs
    assert a.input_units == b.input_units
    assert a.outputs == b.outputs
    assert a.input_units_allow_dimensionless == b.input_units_allow_dimensionless
    assert a.input_units_equivalencies == b.input_units_equivalencies

    np.testing.assert_array_equal(a.parameters, b.parameters)

    assert a._user_bounding_box == b._user_bounding_box
    try:
        a_bounding_box = a.bounding_box
    except NotImplementedError:
        a_bounding_box = None

    try:
        b_bounding_box = b.bounding_box
    except NotImplementedError:
        b_bounding_box = None

    assert a_bounding_box == b_bounding_box

    assert a.fixed == b.fixed
    assert a.bounds == b.bounds

    assert_model_equal(a._user_inverse, b._user_inverse)


def assert_bounding_box_roundtrip(bounding_box, tmp_path, version=None):
    """
    Assert that a bounding_box can be written to an ASDF file and read back
    in without losing any of its essential properties.
    """
    import asdf

    path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile({"bounding_box": bounding_box}, version=version) as af:
        af.write_to(path)

    with asdf.open(path) as af:
        assert bounding_box == af["bounding_box"](bounding_box._model)


def assert_model_roundtrip(model, tmp_path, version=None):
    """
    Assert that a model can be written to an ASDF file and read back
    in without losing any of its essential properties.
    """
    import asdf

    path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile({"model": model}, version=version) as af:
        af.write_to(path)

    with asdf.open(path) as af:
        assert_model_equal(model, af["model"])
        return af["model"]



import asdf
import pytest
from astropy import units as u
from astropy.time import Time, TimeDelta

from asdf_astropy.testing.helpers import assert_time_delta_equal


def create_time_deltas():
    result = [TimeDelta([1, 2] * u.day)]
    result += [TimeDelta(Time.now() - Time.now(), format=format_) for format_ in TimeDelta.FORMATS]
    result += [TimeDelta(0.125 * u.day, scale=scale) for scale in [*list(TimeDelta.SCALES), None]]

    return result


@pytest.mark.parametrize("time_delta", create_time_deltas())
def test_serialization(time_delta, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["time_delta"] = time_delta
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_time_delta_equal(af["time_delta"], time_delta)



from asdf_astropy.converters.helpers import get_tag_name

from .core import TransformConverterBase

__all__ = ["CompoundConverter"]


_OPERATOR_TO_TAG_NAME = {
    "+": "add",
    "-": "subtract",
    "*": "multiply",
    "/": "divide",
    "**": "power",
    "|": "compose",
    "&": "concatenate",
    "fix_inputs": "fix_inputs",
}


_TAG_NAME_TO_MODEL_METHOD = {
    "add": "__add__",
    "subtract": "__sub__",
    "multiply": "__mul__",
    "divide": "__truediv__",
    "power": "__pow__",
    "compose": "__or__",
    "concatenate": "__and__",
    "fix_inputs": "fix_inputs",
}


class CompoundConverter(TransformConverterBase):
    """
    ASDF serialization support for CompoundModel.
    """

    tags = (
        "tag:stsci.edu:asdf/transform/add-*",
        "tag:stsci.edu:asdf/transform/subtract-*",
        "tag:stsci.edu:asdf/transform/multiply-*",
        "tag:stsci.edu:asdf/transform/divide-*",
        "tag:stsci.edu:asdf/transform/power-*",
        "tag:stsci.edu:asdf/transform/compose-*",
        "tag:stsci.edu:asdf/transform/concatenate-*",
        "tag:stsci.edu:asdf/transform/fix_inputs-*",
    )
    types = ("astropy.modeling.core.CompoundModel",)

    def select_tag(self, model, tags, ctx):
        tag_name = _OPERATOR_TO_TAG_NAME[model.op]

        # The extension will never include two tags with the
        # same name but different version, so we can just
        # return the first matching tag that we discover in
        # the list:
        return next(t for t in tags if get_tag_name(t) == tag_name)

    def to_yaml_tree_transform(self, model, tag, ctx):
        left = model.left

        right = (
            {
                "keys": list(model.right.keys()),
                "values": list(model.right.values()),
            }
            if isinstance(model.right, dict)
            else model.right
        )

        return {"forward": [left, right]}

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling.core import CompoundModel, Model

        oper = _TAG_NAME_TO_MODEL_METHOD[get_tag_name(tag)]

        left = node["forward"][0]
        if not isinstance(left, Model):
            msg = f"Unknown left model type '{node['forward'][0]._tag}'"
            raise TypeError(msg)

        right = node["forward"][1]
        if (oper == "fix_inputs" and not isinstance(right, dict)) or (
            oper != "fix_inputs" and not isinstance(right, Model)
        ):
            msg = f"Unknown right model type '{node['forward'][1]._tag}'"
            raise TypeError(msg)

        if oper == "fix_inputs":
            right = dict(zip(right["keys"], right["values"]))
            return CompoundModel("fix_inputs", left, right)

        return getattr(left, oper)(right)



import importlib.resources as importlib_resources

from asdf.resource import DirectoryResourceMapping


def get_resource_mappings():
    """
    Get the resource mapping instances for the astropy schemas
    and manifests.  This method is registered with the
    asdf.resource_mappings entry point.

    Returns
    -------
    list of collections.abc.Mapping
    """
    from . import resources

    resources_root = importlib_resources.files(resources)

    return [
        DirectoryResourceMapping(resources_root / "schemas", "http://astropy.org/schemas/astropy/", recursive=True),
        DirectoryResourceMapping(resources_root / "manifests", "asdf://astropy.org/astropy/manifests/"),
    ]


def get_extensions():
    """
    Get the extension instances for the various astropy
    extensions.  This method is registered with the
    asdf.extensions entry point.

    Returns
    -------
    list of asdf.extension.Extension
    """
    from . import extensions

    return [
        *extensions.ASTROPY_EXTENSIONS,
        extensions.COORDINATES_EXTENSION,
        *extensions.TRANSFORM_EXTENSIONS,
        *extensions.CORE_EXTENSIONS,
    ]



__all__ = [
    "ColumnConverter",
    "AstropyTableConverter",
    "AsdfTableConverter",
    "NdarrayMixinConverter",
]

from .table import AsdfTableConverter, AstropyTableConverter, ColumnConverter, NdarrayMixinConverter



from pathlib import Path

from astropy.tests.runner import TestRunner

__all__ = ["__version__", "test"]

from ._version import version as __version__

# Create the test function for self test
test = TestRunner.make_test_runner_in(Path(__file__).parent)



import asdf
import pytest
from astropy import units


def create_builtin_units():
    return {u for u in list(units.__dict__.values()) if isinstance(u, units.MagUnit)}


@pytest.mark.parametrize("unit", create_builtin_units())
@pytest.mark.filterwarnings("ignore::astropy.units.core.UnitsWarning")
def test_builtin_serialization(unit, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["unit"] = unit
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert af["unit"].is_equivalent(unit)

    with asdf.open(file_path, _force_raw_types=True) as af:
        assert isinstance(af["unit"], asdf.tagged.TaggedDict)
        assert af["unit"]._tag.startswith("tag:astropy.org:astropy/units/magunit-")


def create_magunits():
    magunits = []
    for u in units.__dict__.values():
        if isinstance(u, units.UnitBase) and not isinstance(u, units.MagUnit):
            try:
                magunit = units.mag(u)
            except units.UnitConversionError:
                pass
            else:
                magunits.append(magunit)

    return frozenset(magunits)


@pytest.mark.parametrize("unit", create_magunits())
@pytest.mark.filterwarnings("ignore::astropy.units.core.UnitsWarning")
def test_magunit_serialization(unit, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["unit"] = unit
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert af["unit"].is_equivalent(unit)

    with asdf.open(file_path, _force_raw_types=True) as af:
        assert isinstance(af["unit"], asdf.tagged.TaggedDict)
        assert af["unit"]._tag.startswith("tag:astropy.org:astropy/units/magunit-")



import asdf
import astropy.units as u
import pytest
from astropy.coordinates import Angle, representation
from numpy.random import random

from asdf_astropy.testing.helpers import assert_representation_equal

IGNORED_REPRESENTATION_CLASSES = [
    "WGS84GeodeticRepresentation",
    "WGS72GeodeticRepresentation",
    "GRS80GeodeticRepresentation",
]

REPRESENTATION_CLASSES = [
    getattr(representation, class_name)
    for class_name in representation.__all__
    if "Base" not in class_name and class_name not in IGNORED_REPRESENTATION_CLASSES
]


def create_representation(rep_class):
    kwargs = {}
    for attr_name, attr_type in rep_class.attr_classes.items():
        value = random((100,)) * u.deg if issubclass(attr_type, Angle) else random((100,)) * u.km
        kwargs[attr_name] = value

    return rep_class(**kwargs)


@pytest.mark.parametrize("rep_class", REPRESENTATION_CLASSES)
def test_serialization(rep_class, tmp_path):
    rep = create_representation(rep_class)
    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        af["rep"] = rep
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_representation_equal(af["rep"], rep)



from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType


class QuantityConverter(Converter):
    tags = ("tag:stsci.edu:asdf/unit/quantity-*",)
    types = (
        "astropy.units.quantity.Quantity",
        # The Distance class has no tag of its own, so we
        # just serialize it as a quantity.
        "astropy.coordinates.distances.Distance",
    )

    def to_yaml_tree(self, obj, tag, ctx):
        node = {
            "value": obj.value,
            "unit": obj.unit,
        }

        if obj.isscalar:
            node["datatype"] = obj.dtype.name

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.units import Quantity

        value = node["value"]
        dtype = node.get("datatype", None)
        if isinstance(value, NDArrayType):
            # TODO: Why doesn't NDArrayType work?  This needs some research
            # and documentation.
            value = value._make_array()
            dtype = value.dtype

        return Quantity(value, unit=node["unit"], copy=False, dtype=dtype)



from .core import TransformConverterBase, parameter_to_value


class Rotate3DConverter(TransformConverterBase):
    """
    ASDF support for serializing rotation models that
    use the rotate3d tag.
    """

    tags = ("tag:stsci.edu:asdf/transform/rotate3d-*",)
    types = (
        "astropy.modeling.rotations.RotateNative2Celestial",
        "astropy.modeling.rotations.RotateCelestial2Native",
        "astropy.modeling.rotations.EulerAngleRotation",
    )

    def to_yaml_tree_transform(self, model, tag, ctx):
        from astropy.modeling import rotations

        if isinstance(model, rotations.RotateNative2Celestial):
            node = {
                "phi": parameter_to_value(model.lon),
                "theta": parameter_to_value(model.lat),
                "psi": parameter_to_value(model.lon_pole),
                "direction": "native2celestial",
            }
        elif isinstance(model, rotations.RotateCelestial2Native):
            node = {
                "phi": parameter_to_value(model.lon),
                "theta": parameter_to_value(model.lat),
                "psi": parameter_to_value(model.lon_pole),
                "direction": "celestial2native",
            }
        else:
            node = {
                "phi": parameter_to_value(model.phi),
                "theta": parameter_to_value(model.theta),
                "psi": parameter_to_value(model.psi),
                "direction": model.axes_order,
            }

        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling import rotations

        if node["direction"] == "native2celestial":
            return rotations.RotateNative2Celestial(node["phi"], node["theta"], node["psi"])

        if node["direction"] == "celestial2native":
            return rotations.RotateCelestial2Native(node["phi"], node["theta"], node["psi"])

        return rotations.EulerAngleRotation(node["phi"], node["theta"], node["psi"], axes_order=node["direction"])


class RotationSequenceConverter(TransformConverterBase):
    """
    ASDF support for serializing rotation sequence models.
    """

    tags = ("tag:stsci.edu:asdf/transform/rotate_sequence_3d-*",)
    types = (
        "astropy.modeling.rotations.RotationSequence3D",
        "astropy.modeling.rotations.SphericalRotationSequence",
    )

    def to_yaml_tree_transform(self, model, tag, ctx):
        from astropy.modeling import rotations

        node = {"angles": list(model.angles.value)}
        node["axes_order"] = model.axes_order
        if isinstance(model, rotations.SphericalRotationSequence):
            node["rotation_type"] = "spherical"
        elif isinstance(model, rotations.RotationSequence3D):
            node["rotation_type"] = "cartesian"
        else:
            msg = f"Cannot serialize model of type {type(model)}"
            raise TypeError(msg)
        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling import rotations

        angles = node["angles"]
        axes_order = node["axes_order"]
        rotation_type = node["rotation_type"]
        if rotation_type == "cartesian":
            return rotations.RotationSequence3D(angles, axes_order=axes_order)

        if rotation_type == "spherical":
            return rotations.SphericalRotationSequence(angles, axes_order=axes_order)

        msg = f"Unrecognized rotation_type: {rotation_type}"
        raise ValueError(msg)



__all__ = [
    "CompoundConverter",
    "TransformConverterBase",
    "SimpleTransformConverter",
    "ConstantConverter",
    "IdentityConverter",
    "RemapAxesConverter",
    "UnitsMappingConverter",
    "MathFunctionsConverter",
    "PolynomialConverter",
    "OrthoPolynomialConverter",
    "ProjectionConverter",
    "ModelBoundingBoxConverter",
    "CompoundBoundingBoxConverter",
    "Rotate3DConverter",
    "RotationSequenceConverter",
    "SplineConverter",
    "TabularConverter",
]

from .compound import CompoundConverter
from .core import SimpleTransformConverter, TransformConverterBase
from .functional_models import ConstantConverter
from .mappings import IdentityConverter, RemapAxesConverter, UnitsMappingConverter
from .math_functions import MathFunctionsConverter
from .polynomial import OrthoPolynomialConverter, PolynomialConverter
from .projections import ProjectionConverter
from .properties import CompoundBoundingBoxConverter, ModelBoundingBoxConverter
from .rotations import Rotate3DConverter, RotationSequenceConverter
from .spline import SplineConverter
from .tabular import TabularConverter



__all__ = [
    "AngleConverter",
    "LatitudeConverter",
    "LongitudeConverter",
    "EarthLocationConverter",
    "FrameConverter",
    "LegacyICRSConverter",
    "RepresentationConverter",
    "SkyCoordConverter",
    "SpectralCoordConverter",
    "EquivalencyConverter",
    "QuantityConverter",
    "UnitConverter",
    "MagUnitConverter",
    "FitsConverter",
    "AsdfFitsConverter",
    "AstropyFitsConverter",
    "ColumnConverter",
    "AstropyTableConverter",
    "AsdfTableConverter",
    "NdarrayMixinConverter",
    "TimeDeltaConverter",
    "TimeConverter",
    "CompoundConverter",
    "TransformConverterBase",
    "SimpleTransformConverter",
    "ConstantConverter",
    "IdentityConverter",
    "RemapAxesConverter",
    "UnitsMappingConverter",
    "MathFunctionsConverter",
    "PolynomialConverter",
    "OrthoPolynomialConverter",
    "ProjectionConverter",
    "ModelBoundingBoxConverter",
    "CompoundBoundingBoxConverter",
    "Rotate3DConverter",
    "RotationSequenceConverter",
    "SplineConverter",
    "TabularConverter",
]

from .coordinates import (
    AngleConverter,
    EarthLocationConverter,
    FrameConverter,
    LatitudeConverter,
    LegacyICRSConverter,
    LongitudeConverter,
    RepresentationConverter,
    SkyCoordConverter,
    SpectralCoordConverter,
)
from .fits import AsdfFitsConverter, AstropyFitsConverter, FitsConverter
from .table import AsdfTableConverter, AstropyTableConverter, ColumnConverter, NdarrayMixinConverter
from .time import TimeConverter, TimeDeltaConverter
from .transform import (
    CompoundBoundingBoxConverter,
    CompoundConverter,
    ConstantConverter,
    IdentityConverter,
    MathFunctionsConverter,
    ModelBoundingBoxConverter,
    OrthoPolynomialConverter,
    PolynomialConverter,
    ProjectionConverter,
    RemapAxesConverter,
    Rotate3DConverter,
    RotationSequenceConverter,
    SimpleTransformConverter,
    SplineConverter,
    TabularConverter,
    TransformConverterBase,
    UnitsMappingConverter,
)
from .unit import EquivalencyConverter, MagUnitConverter, QuantityConverter, UnitConverter



from .core import TransformConverterBase


class SplineConverter(TransformConverterBase):
    """
    ASDF support for serializing 1D spline models
    """

    tags = ("tag:stsci.edu:asdf/transform/spline1d-*",)
    types = ("astropy.modeling.spline.Spline1D",)

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {"knots": model.t, "coefficients": model.c, "degree": model.degree}

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling.spline import Spline1D

        knots = node["knots"]
        coeffs = node["coefficients"]
        degree = node["degree"]

        return Spline1D(knots, coeffs, degree)



from asdf.extension import Converter

from asdf_astropy.converters.utils import import_type


class FrameConverter(Converter):
    def __init__(self, tags, frame_type_name):
        self._frame_type_name = frame_type_name
        self._frame_type = None

        if isinstance(tags, str):
            tags = [tags]
        self._tags = tags

    @property
    def tags(self):
        return self._tags

    @property
    def types(self):
        return [self._frame_type_name]

    @property
    def frame_type(self):
        # Delay import until the frame class is needed to improve speed
        # of loading the extension.
        if self._frame_type is None:
            self._frame_type = import_type(self._frame_type_name)
        return self._frame_type

    def to_yaml_tree(self, obj, tag, ctx):
        node = {}

        if obj.has_data:
            node["data"] = obj.data

        # TODO: Figure out why we can't use the frame_attributes
        # values and document.
        frame_attributes = {}
        for attr in obj.frame_attributes:
            value = getattr(obj, attr, None)
            if value is not None:
                frame_attributes[attr] = value
        node["frame_attributes"] = frame_attributes

        return node

    def from_yaml_tree(self, node, tag, ctx):
        data = node.get("data", None)
        if data is not None:
            return self.frame_type(node["data"], **node["frame_attributes"])

        return self.frame_type(**node["frame_attributes"])


class LegacyICRSConverter(Converter):
    tags = ("tag:astropy.org:astropy/coordinates/frames/icrs-1.0.0",)
    # Leave the types list empty so that the 1.1.0 ICRS converter
    # is used on write.
    types = ()

    def to_yaml_tree(self, obj, tag, ctx):
        from astropy.units import Quantity

        return {
            "ra": {
                "value": obj.ra.value,
                "unit": obj.ra.unit.to_string(),
                "wrap_angle": Quantity(obj.ra.wrap_angle),
            },
            "dec": {
                "value": obj.dec.value,
                "unit": obj.dec.unit.to_string(),
            },
        }

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates import ICRS, Angle, Latitude, Longitude

        ra = Longitude(node["ra"]["value"], unit=node["ra"]["unit"], wrap_angle=Angle(node["ra"]["wrap_angle"]))

        dec = Latitude(node["dec"]["value"], unit=node["dec"]["unit"])

        return ICRS(ra=ra, dec=dec)



import warnings

from asdf.extension import Converter


class UnitConverter(Converter):
    tags = (
        "tag:stsci.edu:asdf/unit/unit-*",
        "tag:astropy.org:astropy/units/unit-*",
    )
    types = (
        "astropy.units.core.CompositeUnit",
        "astropy.units.core.IrreducibleUnit",
        "astropy.units.core.NamedUnit",
        "astropy.units.core.PrefixUnit",
        "astropy.units.core.Unit",
        "astropy.units.core.UnitBase",
        "astropy.units.core.UnrecognizedUnit",
        "astropy.units.function.mixin.IrreducibleFunctionUnit",
        "astropy.units.function.mixin.RegularFunctionUnit",
    )

    def select_tag(self, obj, tags, ctx):
        from astropy.units import UnitsError, UnitsWarning

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UnitsWarning)

            try:
                obj.to_string(format="vounit")
            except (UnitsError, ValueError):
                return next(t for t in tags if "astropy.org" in t)

            return next(t for t in tags if "stsci.edu" in t)

    def to_yaml_tree(self, obj, tag, ctx):
        if "stsci.edu" in tag:
            return obj.to_string(format="vounit")

        return obj.to_string()

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.units import Unit

        kwargs = {"parse_strict": "silent"}
        if "stsci.edu" in tag:
            kwargs["format"] = "vounit"

        return Unit(node, **kwargs)



from asdf.extension import Converter


class EquivalencyConverter(Converter):
    tags = ("tag:astropy.org:astropy/units/equivalency-*",)
    types = ("astropy.units.equivalencies.Equivalency",)

    def to_yaml_tree(self, obj, tag, ctx):
        return [
            {"name": name, "kwargs_names": list(kw.keys()), "kwargs_values": list(kw.values())}
            for name, kw in zip(obj.name, obj.kwargs)
        ]

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.cosmology.units import with_H0
        from astropy.units import equivalencies

        components = []
        for equivalency_node in node:
            name = equivalency_node["name"]
            equivalency_method = with_H0 if name == "with_H0" else getattr(equivalencies, name)

            kwargs = dict(zip(equivalency_node["kwargs_names"], equivalency_node["kwargs_values"]))
            components.append(equivalency_method(**kwargs))

        # The Equivalency class is a UserList that overrides __add__ to
        # provide special behavior when combined with another Equivalency.
        # We're using sum here to add each subsequent Equivalency to the
        # first so we end up with a single correctly combined Equivalency
        # object at the end.
        return sum(components[1:], components[0])



from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType


class SpectralCoordConverter(Converter):
    tags = ("tag:astropy.org:astropy/coordinates/spectralcoord-*",)
    types = ("astropy.coordinates.spectral_coordinate.SpectralCoord",)

    def to_yaml_tree(self, obj, tag, ctx):
        node = {
            "value": obj.value,
            "unit": obj.unit,
        }

        if obj.observer is not None:
            node["observer"] = obj.observer

        if obj.target is not None:
            node["target"] = obj.target

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.spectral_coordinate import SpectralCoord

        value = node["value"]
        if isinstance(value, NDArrayType):
            # TODO: Why doesn't NDArrayType work?  This needs some research
            # and documentation.  See similar note in QuantityConverter.
            value = value._make_array()

        return SpectralCoord(
            value,
            unit=node["unit"],
            observer=node.get("observer"),
            target=node.get("target"),
        )



import re
from itertools import chain

from asdf import extension


class CompoundManifestExtension(extension.Extension):
    """
    Combine a listed of asdf ``ManifestExtensions`` into a single extension.
    """

    def __init__(self, extensions):
        self._extensions = extensions
        # overwrite extension uri from first extension (the one from asdf-standard)
        # so that this extension uses a new, unique uri
        self._extension_uri = re.sub("asdf-format", "astropy", extensions[0].extension_uri)

    @property
    def extension_uri(self):
        return self._extension_uri

    @property
    def asdf_standard_requirement(self):
        return self._extensions[0].asdf_standard_requirement

    @property
    def legacy_class_names(self):
        return list(chain.from_iterable(e.legacy_class_names for e in self._extensions))

    @property
    def converters(self):
        return list(chain.from_iterable(e.converters for e in self._extensions))

    @property
    def compressors(self):
        return list(chain.from_iterable(e.compressors for e in self._extensions))

    @property
    def tags(self):
        return list(chain.from_iterable(e.tags for e in self._extensions))



__all__ = [
    "AngleConverter",
    "LatitudeConverter",
    "LongitudeConverter",
    "EarthLocationConverter",
    "FrameConverter",
    "LegacyICRSConverter",
    "RepresentationConverter",
    "SkyCoordConverter",
    "SpectralCoordConverter",
]

from .angle import AngleConverter, LatitudeConverter, LongitudeConverter
from .earth_location import EarthLocationConverter
from .frame import FrameConverter, LegacyICRSConverter
from .representation import RepresentationConverter
from .sky_coord import SkyCoordConverter
from .spectral_coord import SpectralCoordConverter
